# -*- coding: utf-8 -*-

#     Copyright 2025. ThingsBoard
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

from setuptools import setup
from os import path

from thingsboard_gateway import version

current_directory = path.abspath(path.dirname(__file__))
with open(path.join(current_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    version=version.VERSION,
    name="thingsboard-gateway",
    author="ThingsBoard",
    author_email="info@thingsboard.io",
    license="Apache Software License (Apache Software License 2.0)",
    description="Thingsboard Gateway for IoT devices.",
    url="https://github.com/thingsboard/thingsboard-gateway",
    long_description=long_description,
    long_description_content_type="text/markdown",
    include_package_data=True,
    python_requires=">=3.10",
    packages=['thingsboard_gateway', 'thingsboard_gateway.gateway',
              'thingsboard_gateway.gateway.entities',
              'thingsboard_gateway.gateway.proto', 'thingsboard_gateway.gateway.grpc_service',
              'thingsboard_gateway.gateway.shell', 'thingsboard_gateway.gateway.statistics',
              'thingsboard_gateway.storage', 'thingsboard_gateway.storage.memory',
              'thingsboard_gateway.gateway.report_strategy', 'thingsboard_gateway.storage.file',
              'thingsboard_gateway.storage.sqlite',
              'thingsboard_gateway.connectors',
              'thingsboard_gateway.connectors.ble', 'thingsboard_gateway.extensions.ble',
              'thingsboard_gateway.connectors.socket', 'thingsboard_gateway.extensions.socket',
              'thingsboard_gateway.connectors.mqtt', 'thingsboard_gateway.extensions.mqtt',
              'thingsboard_gateway.connectors.xmpp', 'thingsboard_gateway.extensions.xmpp',
              'thingsboard_gateway.connectors.modbus', 'thingsboard_gateway.connectors.modbus.entities',
              'thingsboard_gateway.extensions.modbus',
              'thingsboard_gateway.connectors.opcua', 'thingsboard_gateway.extensions.opcua',
              'thingsboard_gateway.connectors.opcua.entities',
              'thingsboard_gateway.connectors.request', 'thingsboard_gateway.extensions.request',
              'thingsboard_gateway.connectors.ocpp', 'thingsboard_gateway.extensions.ocpp',
              'thingsboard_gateway.connectors.can', 'thingsboard_gateway.extensions.can',
              'thingsboard_gateway.connectors.odbc', 'thingsboard_gateway.extensions.odbc',
              'thingsboard_gateway.connectors.bacnet', 'thingsboard_gateway.connectors.bacnet.entities',
              'thingsboard_gateway.extensions.bacnet',
              'thingsboard_gateway.connectors.rest', 'thingsboard_gateway.extensions.rest',
              'thingsboard_gateway.connectors.snmp', 'thingsboard_gateway.extensions.snmp',
              'thingsboard_gateway.connectors.ftp', 'thingsboard_gateway.extensions.ftp',
              'thingsboard_gateway.connectors.knx', 'thingsboard_gateway.extensions.knx',
              'thingsboard_gateway.connectors.knx.entities',
              'thingsboard_gateway.tb_utility', 'thingsboard_gateway.extensions',
              'thingsboard_gateway.extensions.serial'
              ],
    install_requires=[
        'setuptools',
        'cryptography',
        'jsonpath-rw',
        'regex',
        'pip',
        'PyYAML',
        'orjson',
        'pybase64',
        'simplejson',
        'urllib3>=2.3.0',
        'requests>=2.32.3',
        'mmh3',
        'grpcio',
        'protobuf',
        'python-dateutil',
        'cachetools',
        'tb-paho-mqtt-client>=2.1.2',
        'tb-mqtt-client==1.13.9',
        'packaging==23.1',
        'service-identity',
        'psutil',
        'PySocks',
    ],
    download_url='https://github.com/thingsboard/thingsboard-gateway/archive/%s.tar.gz' % version.VERSION,
    entry_points={
        'console_scripts': [
            'thingsboard-gateway = thingsboard_gateway.tb_gateway:daemon',
            'tb-gateway-shell = thingsboard_gateway.gateway.shell:main'
        ]
    })



#     Copyright 2025. ThingsBoard
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.


from typing import Any, Tuple
from simplejson import loads

from thingsboard_gateway.connectors.converter import Converter
from thingsboard_gateway.gateway.constants import REPORT_STRATEGY_PARAMETER, TELEMETRY_PARAMETER, TIMESERIES_PARAMETER
from thingsboard_gateway.gateway.entities.converted_data import ConvertedData
from thingsboard_gateway.gateway.entities.datapoint_key import DatapointKey
from thingsboard_gateway.gateway.entities.report_strategy_config import ReportStrategyConfig
from thingsboard_gateway.gateway.entities.telemetry_entry import TelemetryEntry
from thingsboard_gateway.tb_utility.tb_utility import TBUtility


class SerialUplinkConverter(Converter):
    """
    Uplink converter is used to convert incoming data to the format that platform expects.
    Such as we create uplink converter for each configured device,
    this converter is used to convert incoming data from only one device.
    Because config, that we passed to init method, is device specific.
    If your connector can handle multiple devices, you can create one converter for all devices.
    """

    def __init__(self, config, logger):
        self._log = logger
        self.__config = config
        self.__device_report_strategy = None
        self.__device_name = self.__config.get('deviceName', self.__config.get('name', 'SerialDevice'))
        self.__device_type = self.__config.get('deviceType', self.__config.get('type', 'default'))
        try:
            self.__device_report_strategy = ReportStrategyConfig(self.__config.get(REPORT_STRATEGY_PARAMETER))
        except ValueError as e:
            self._log.trace("Report strategy config is not specified for device %s: %s", self.__device_name, e)

    def convert(self, config, data: bytes):
        """Converts incoming data to the format that platform expects. Config is specified only for RPC responses."""
        self._log.debug("Data to convert: %s", data)
        if config is not None:
            converted_data = {"result": self.__convert_value_to_type(data, config)}
            return converted_data
        else:
            converted_data = ConvertedData(self.__device_name, self.__device_type)
            for datapoint_config in self.__config.get(TIMESERIES_PARAMETER, self.__config.get(TELEMETRY_PARAMETER, [])):
                try:
                    telemetry_entry = self.__convert_telemetry_datapoint(data, datapoint_config)
                    if telemetry_entry:
                        converted_data.add_to_telemetry(telemetry_entry)
                except Exception as e:
                    self._log.error("Error converting telemetry datapoint: %s", e)
            for datapoint_config in self.__config.get('attributes', []):
                try:
                    attribute_data = self.__convert_attributes_datapoint(data, datapoint_config)
                    if attribute_data:
                        converted_data.add_to_attributes(*attribute_data)
                except Exception as e:
                    self._log.error("Error converting attribute datapoint: %s", e)
            self._log.debug("Converted data: %s", converted_data)
        return converted_data

    def __convert_telemetry_datapoint(self, data, dp_config) -> TelemetryEntry:
        key = dp_config.get('key')
        datapoint_key = self.__convert_datapoint_key(key, dp_config, self.__device_report_strategy, self._log)
        value = self.__convert_value_to_type(data, dp_config)
        if not datapoint_key or not value:
            self._log.trace("Datapoint %s - not found in incoming data: %s", key, data.hex())
            return None
        return TelemetryEntry({datapoint_key: value})

    def __convert_attributes_datapoint(self, data, dp_config) -> Tuple[DatapointKey, Any]:
        key = dp_config.get('key')
        datapoint_key = self.__convert_datapoint_key(key, dp_config, self.__device_report_strategy, self._log)
        value = self.__convert_value_to_type(data, dp_config)
        if not datapoint_key or not value:
            self._log.trace("Datapoint %s - not found in incoming data: %s", key, data.hex())
            return None
        return (datapoint_key, value)

    @staticmethod
    def __convert_value_to_type(data, dp_config):
        type_ = dp_config.get('type')
        data_for_conversion = data
        if dp_config.get("untilDelimiter") or dp_config.get("fromDelimiter"):
            fromDelimiter = dp_config.get("fromDelimiter")
            untilDelimiter = dp_config.get("untilDelimiter")
            fromDelimiterPosition = data_for_conversion.find(
                fromDelimiter.encode('UTF-8')) if fromDelimiter else 0
            untilDelimiterPosition = data_for_conversion.find(
                untilDelimiter.encode('UTF-8')) if untilDelimiter else -1
            if fromDelimiterPosition != -1 \
                    and untilDelimiterPosition != -1 \
                    and fromDelimiterPosition < untilDelimiterPosition:
                data_for_conversion = data_for_conversion[fromDelimiterPosition:untilDelimiterPosition]
            elif fromDelimiterPosition != -1 and fromDelimiterPosition < len(data_for_conversion):
                data_for_conversion = data_for_conversion[fromDelimiterPosition:]
            elif untilDelimiterPosition != -1 and untilDelimiterPosition < len(data_for_conversion):
                data_for_conversion = data_for_conversion[:untilDelimiterPosition]
        elif dp_config.get("fromByte") or dp_config.get("toByte"):
            if dp_config.get("fromByte") and dp_config.get("toByte") \
                    and dp_config["fromByte"] < dp_config["toByte"] \
                    and len(data_for_conversion) > dp_config["toByte"]:
                data_for_conversion = data_for_conversion[dp_config["toByte"]:dp_config["fromByte"]]
            else:
                if dp_config.get("fromByte") and len(data_for_conversion) > dp_config.get("fromByte", 0):
                    data_for_conversion = data_for_conversion[dp_config["fromByte"]:]
                if dp_config.get("toByte") and \
                        (len(data_for_conversion) > dp_config.get("toByte", 0) or dp_config["toByte"] == -1):
                    data_for_conversion = data_for_conversion[:dp_config["toByte"]]

        if type_ == 'string':
            value = data_for_conversion.decode('UTF-8').strip()
        elif type_ == 'json':
            value = loads(data_for_conversion.decode('UTF-8'))
        elif type_ == 'int':
            value = int(data_for_conversion)
        elif type_ == 'float' or type_ == 'double':
            value = float(data_for_conversion)
        elif type_ == 'bool':
            try:
                value = bool(int(data_for_conversion))
            except ValueError:
                return data_for_conversion.decode('UTF-8').strip().lower() == 'true'
        else:
            value = data_for_conversion.hex()
        return value

    @staticmethod
    def __convert_datapoint_key(key, dp_config, device_report_strategy, logger):
        return TBUtility.convert_key_to_datapoint_key(key, device_report_strategy, dp_config, logger)



#     Copyright 2025. ThingsBoard
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

import inspect
import os.path
from copy import deepcopy
from logging import getLogger
from logging.config import dictConfig
from os.path import exists
from queue import Queue, Empty
from threading import Thread
from time import sleep, time, monotonic
from typing import TYPE_CHECKING

from regex import fullmatch
from simplejson import dumps, load

from tb_gateway_mqtt import TBGatewayMqttClient

from thingsboard_gateway.gateway.constants import CONFIG_VERSION_PARAMETER, REPORT_STRATEGY_PARAMETER, \
    DEFAULT_REPORT_STRATEGY_CONFIG, CONNECTOR_PARAMETER, ReportStrategy
from thingsboard_gateway.gateway.entities.report_strategy_config import ReportStrategyConfig
from thingsboard_gateway.gateway.report_strategy.report_strategy_service import ReportStrategyService
from thingsboard_gateway.gateway.tb_client import TBClient

if TYPE_CHECKING:
    from thingsboard_gateway.gateway.tb_gateway_service import TBGatewayService
from thingsboard_gateway.tb_utility.tb_logger import TbLogger
from thingsboard_gateway.tb_utility.tb_utility import TBUtility


class RemoteConfigurator:
    ALLOWED_BACKUPS_AMOUNT = 10
    RECEIVED_UPDATE_QUEUE = Queue()

    DEFAULT_STATISTICS = {
        'enable': True,
        'enableCustom': True,
        'statsSendPeriodInSeconds': 60,
        'customStatsSendPeriodInSeconds': 300,
        'configuration': 'statistics.json'
    }

    def __init__(self, gateway: 'TBGatewayService', config):
        self.__log = getLogger("service")
        self._request_queue = Queue()
        self.in_process = False
        self._gateway = gateway
        self._config = config
        self._load_connectors_configuration()
        self._logs_configuration = self._load_logs_configuration()
        self._active_connectors = []
        self._handlers = {
            'storage_configuration': self._handle_storage_configuration_update,
            'grpc_configuration': self._handle_grpc_configuration_update,
            'logs_configuration': self._handle_logs_configuration_update,
            'active_connectors': self._handle_active_connectors_update,
            'RemoteLoggingLevel': self._handle_remote_logging_level_update,
            r'(?=\D*\d?).*': self._handle_connector_configuration_update,
        }
        self._modifiable_static_attrs = {
            'logs_configuration': 'logs.json'
        }
        self._max_backup_files_number = 10

        # creating backups (general and connectors)
        self.create_configuration_file_backup(self._get_general_config_in_local_format(), "tb_gateway.json")
        self._create_connectors_backup()

        self._requests_processing_thread = Thread(name='Remote Request Processing', target=self._process_config_request,
                                                  daemon=True)
        self._requests_processing_thread.start()

        self.__log.info('Remote Configurator started')

    @property
    def general_configuration(self):
        return self._config.get('thingsboard', {})

    @general_configuration.setter
    def general_configuration(self, config):
        self._config['thingsboard'].update(config)

    @property
    def storage_configuration(self):
        return self._config.get('storage', {})

    @storage_configuration.setter
    def storage_configuration(self, config):
        self._config['storage'].update(config)

    @property
    def grpc_configuration(self):
        return self._config.get('grpc', {})

    @grpc_configuration.setter
    def grpc_configuration(self, config):
        self._config.get('grpc', {}).update(config)

    @property
    def connectors_configuration(self):
        connectors = self._config.get('connectors', [])
        for connector in connectors:
            connector.pop('config_updated', None)
            connector.pop('config_file_path', None)
        return connectors

    def _get_active_connectors(self):
        return [connector['name'] for connector in self.connectors_configuration]

    def _get_general_config_in_local_format(self):
        """
        Method returns general configuration in format that should be used only for local files
        !!!Don't use it for sending data to TB (use `_get_general_config_in_remote_format` instead)!!!
        """
        connectors_config = []
        for connector in self.connectors_configuration:
            config = {
                'type': connector['type'],
                'name': connector['name'],
                'configuration': connector['configuration']
            }
            if connector.get('class') is not None:
                config['class'] = connector.get('class')
            connectors_config.append(config)

        adopted_general_config = {}
        adopted_general_config.update(self.general_configuration)

        security_section = adopted_general_config.get('security')
        security_type = security_section.get('type', 'accessToken')
        if security_type == 'accessToken':
            security_section.pop('clientId', None)
            security_section.pop('username', None)
            security_section.pop('password', None)
        elif security_type == 'usernamePassword':
            security_section.pop('accessToken', None)

        return {
            'thingsboard': adopted_general_config,
            'storage': self.storage_configuration,
            'grpc': self.grpc_configuration,
            'connectors': connectors_config
        }

    def _get_general_config_in_remote_format(self):
        """
        Method returns general configuration in format that should be used only for sending configuration to the server.
        !!!Don't use it for saving data to conf files (use `_get_general_config_in_local_format`)!!!
        """

        config = deepcopy(self.general_configuration)
        stat_conf_path = config.get('statistics', {}).get('configuration', 'statistics.json')
        commands = []
        if stat_conf_path and exists(self._gateway.get_config_path() + stat_conf_path):
            with open(self._gateway.get_config_path() + stat_conf_path, 'r') as file:
                commands = load(file)
        else:
            with open(self._gateway.get_config_path() + 'statistics.json', 'w') as file:
                file.writelines(dumps(commands, indent='  '))
        statistics_config = self.general_configuration.get('statistics', {})
        enable_custom = ((statistics_config.get('enable', False) and statistics_config.get('enableCustom') is None)
                         or statistics_config.get('enableCustom', False))
        config['statistics'] = {
                'enable': self.general_configuration.get('statistics', {}).get('enable', False),
                'enableCustom': enable_custom,
                'statsSendPeriodInSeconds': self.general_configuration.get('statistics', {}).get('statsSendPeriodInSeconds', 60), # noqa
                'customStatsSendPeriodInSeconds': self.general_configuration.get('statistics', {}).get('customStatsSendPeriodInSeconds', 300), # noqa
                'configuration': self.general_configuration['statistics'].get('configuration', 'statistics.json'),
                'commands': commands
        }
        return config

    def send_current_configuration(self):
        """
        Calling manually only on RemoteConfigurator init (TbGatewayService.__init_remote_configuration method)
        When Gateway started, sending all configs for states synchronizing
        """

        self.__log.debug('Sending all configurations (init)')
        ts = int(time() * 1000)
        init_config_message = {
            'general_configuration': {**self._get_general_config_in_remote_format(), 'ts': ts},
            'storage_configuration': {**self.storage_configuration, 'ts': ts},
            'grpc_configuration': {**self.grpc_configuration, 'ts': ts},
            'logs_configuration': {**self._logs_configuration, 'ts': ts},
            'active_connectors': self._get_active_connectors(),
            'Version': self._gateway.version.get('current_version', '0.0')
        }
        self._gateway.send_attributes(init_config_message)

        # sending remote created connectors
        for connector in self.connectors_configuration:
            self.send_connector_current_configuration(connector)

    def send_connector_current_configuration(self, connector_configuration: dict):
        config_to_send = {**connector_configuration,
                          'logLevel': connector_configuration.get('configurationJson', {}).get('logLevel', 'INFO'),
                          'enableRemoteLogging':
                              connector_configuration.get('configurationJson', {}).get('enableRemoteLogging', False),
                          'ts': int(time() * 1000)
                          }
        if config_to_send.get('config') is not None:
            config_to_send.pop('config', None)
        if config_to_send.get("configurationJson", {}) is not None:
            config_to_send.get("configurationJson", {}).pop('logLevel', None)
            config_to_send.get("configurationJson", {}).pop('enableRemoteLogging', None)
        if config_to_send.get('configurationJson', {}).get(REPORT_STRATEGY_PARAMETER) is not None:
            config_to_send[REPORT_STRATEGY_PARAMETER] = config_to_send['configurationJson'].pop(REPORT_STRATEGY_PARAMETER) # noqa
        if config_to_send.get('configurationJson', {}).get(CONFIG_VERSION_PARAMETER) is not None:
            config_to_send[CONFIG_VERSION_PARAMETER] = config_to_send['configurationJson'].pop(CONFIG_VERSION_PARAMETER)
        self._gateway.send_attributes({connector_configuration['name']: config_to_send})

    def _load_connectors_configuration(self):
        for (_, connector_list) in self._gateway.connectors_configs.items():
            for connector in connector_list:
                for general_connector_config in self.connectors_configuration:
                    if general_connector_config['name'] == connector['name']:
                        config = connector.get('config')[general_connector_config['configuration']]
                        general_connector_config.update(connector)
                        general_connector_config['configurationJson'] = config

    def _load_logs_configuration(self):
        """
        Calling only on RemoteConfigurator init (__init__ method)
        """

        try:
            with open(self._gateway.get_config_path() + 'logs.json', 'r') as logs:
                return load(logs)
        except Exception:
            self.__log.warning("Cannot open logs configuration file. Using default logs configuration.")
            return {}

    def _process_config_request(self):
        while not self._gateway.stopped:
            try:
                if not RemoteConfigurator.RECEIVED_UPDATE_QUEUE.empty():
                    self.in_process = True
                    config = RemoteConfigurator.RECEIVED_UPDATE_QUEUE.get(timeout=.1)
                    if len(config) > 1:
                        self.__log.info('Requested remote configuration received, keys: %s', list(config.keys()))
                    else:
                        self.__log.info('Remote configuration received, key: %s', list(config.keys()))
                    self.__log.debug('Got config update request: %s', config)

                    try:
                        if 'general_configuration' in config.keys():
                            self._handle_general_configuration_update(config['general_configuration'])
                            config.pop('general_configuration', None)

                        for attr_name in config.keys():
                            if 'deleted' in attr_name:
                                continue

                            request_config = config[attr_name]
                            if not self._is_modified(attr_name, request_config) and self.__is_running(request_config):
                                continue

                            request_processed = False
                            for (name, func) in self._handlers.items():
                                if fullmatch(name, attr_name):
                                    func(request_config)
                                    request_processed = True
                                    break
                            if not request_processed:
                                self.__log.error("Cannot process request for %s", attr_name)
                    except (KeyError, AttributeError) as e:
                        self.__log.error('Unknown attribute update name (Available: %s), %r',
                                         list(self._handlers.keys()),
                                         e)

                    self.in_process = False
                else:
                    self._gateway.stop_event.wait(.5)
            except (TimeoutError, Empty):
                self._gateway.stop_event.wait(.5)
                pass
            except Exception as e:
                self.__log.error('Exception while processing configuration update request.', exc_info=e)

    # HANDLERS ---------------------------------------------------------------------------------------------------------
    def _handle_general_configuration_update(self, config):
        """
        General configuration update handling in 5 steps:
        1. Checking if connection changed (host, port, QoS, security or provisioning sections);
        2. Checking if statistics collecting changed;
        3. Checking if device filtering changed;
        4. Checking if Remote Shell on/off state changed;
        5. Updating other params (regardless of whether they have changed):
            a. maxPayloadSizeBytes;
            b. minPackSendDelayMS;
            c. minPackSizeToSend;
            d. checkConnectorsConfigurationInSeconds;
            f. handleDeviceRenaming.

        If config from steps 1-4 changed:
        True -> applying new config with related objects creating (old objects will remove).
        """

        self.__log.debug('Processing general configuration update')

        self.__log.debug('--- Checking connection configuration changes...')
        security_match = self.__is_security_match(self.general_configuration.get('security'),
                                                  config.get('security', {}))

        if (config['host'] != self.general_configuration['host']
                or config['port'] != self.general_configuration['port']
                or not security_match
                or config.get('provisioning', {}) != self.general_configuration.get('provisioning', {})
                or config['qos'] != self.general_configuration['qos']):
            self.__log.debug('---- Connection configuration changed. Processing...')
            success = self._apply_connection_config(config)
            if self._gateway.stopped:
                return
            if not success:
                config.update(self.general_configuration)
        else:
            self.__log.debug('--- Connection configuration not changed.')

        self.__log.debug('--- Checking statistics configuration changes...')
        changed = self._check_statistics_configuration_changes(config.get('statistics', self.DEFAULT_STATISTICS))
        if changed:
            self.__log.debug('---- Statistics configuration changed. Processing...')
            success = self._apply_statistics_config(config.get('statistics', self.DEFAULT_STATISTICS))
            if not success:
                config['statistics'].update(self.general_configuration['statistics'])
        else:
            self.__log.debug('--- Statistics configuration not changed.')

        self.__log.debug('--- Checking device filtering configuration changes...')
        if config.get('deviceFiltering') != self.general_configuration.get('deviceFiltering'):
            self.__log.debug('---- Device filtering configuration changed. Processing...')
            success = self._apply_device_filtering_config(config)
            if not success:
                config['deviceFiltering'].update(self.general_configuration['deviceFiltering'])
        else:
            self.__log.debug('--- Device filtering configuration not changed.')

        self.__log.debug('--- Checking Remote Shell configuration changes...')
        if config.get('remoteShell') != self.general_configuration.get('remoteShell'):
            self.__log.debug('---- Remote Shell configuration changed. Processing...')
            success = self._apply_remote_shell_config(config)
            if not success:
                config['remoteShell'].update(self.general_configuration['remoteShell'])
        else:
            self.__log.debug('--- Remote Shell configuration not changed.')

        self.__log.debug('--- Checking other configuration parameters changes...')

        self.__log.debug('--- Checking Report Strategy configuration changes...')
        if config.get(REPORT_STRATEGY_PARAMETER) != self.general_configuration.get(REPORT_STRATEGY_PARAMETER):
            self.__log.debug('---- Report Strategy configuration changed. Processing...')
            success = self._apply_report_strategy_config(config)
            if success:
                if REPORT_STRATEGY_PARAMETER not in config:
                    config[REPORT_STRATEGY_PARAMETER] = DEFAULT_REPORT_STRATEGY_CONFIG
                self.general_configuration[REPORT_STRATEGY_PARAMETER].update(config[REPORT_STRATEGY_PARAMETER])
            else:
                config[REPORT_STRATEGY_PARAMETER].update(self.general_configuration[REPORT_STRATEGY_PARAMETER])
        else:
            self.__log.debug('--- Report Strategy configuration not changed.')

        self.__log.debug('--- Checking other configuration parameters changes...')
        self._apply_other_params_config(config)

        self.__log.debug('--- Saving new general configuration...')
        self.general_configuration = config
        self._gateway.send_attributes({'general_configuration': self.general_configuration})
        self._cleanup()
        with open(self._gateway.get_config_path() + "tb_gateway.json", "w",  encoding="UTF-8") as file:
            file.writelines(dumps(self._get_general_config_in_local_format(), indent='  '))

    def _handle_storage_configuration_update(self, config):
        self.__log.debug('Processing storage configuration update...')

        old_event_storage_config = self._gateway._event_storage.get_configuration()
        self._gateway._event_storage.stop()
        storage_logger = self._gateway.remote_handler.get_logger('storage')
        try:
            storage_class = self._gateway.event_storage_types[config["type"]]
            self._gateway._event_storage.stop()
            self._gateway._event_storage = storage_class(config, storage_logger, self._gateway.stop_event)
        except Exception as e:
            self.__log.error('Something went wrong with applying the new storage configuration. Reverting...')
            self.__log.exception(e)
            old_storage_class = self._gateway.event_storage_types[old_event_storage_config["type"]]
            self._gateway._event_storage = old_storage_class(old_event_storage_config,
                                                             storage_logger,
                                                             self._gateway.stop_event)
        else:
            self.storage_configuration = config
            with open(self._gateway.get_config_path() + "tb_gateway.json", "w", encoding="UTF-8") as file:
                file.writelines(dumps(self._get_general_config_in_local_format(), indent='  '))
            self._gateway.send_attributes({'storage_configuration': self.storage_configuration})

            self.__log.debug('Processed storage configuration update successfully')

    def _handle_grpc_configuration_update(self, config):
        self.__log.debug('Processing GRPC configuration update...')
        if config.get('enabled', False):
            try:
                self._gateway.init_grpc_service(config)
                for connector_name in self._gateway.available_connectors_by_name:
                    self._gateway.available_connectors_by_name[connector_name].close()
                self._gateway.load_connectors(self._get_general_config_in_local_format())
                self._gateway.connect_with_connectors()
            except Exception as e:
                self.__log.error('Something went wrong with applying the new GRPC configuration. Reverting...')
                self.__log.exception(e)
                self._gateway.init_grpc_service(self.grpc_configuration)
                for connector_name in self._gateway.available_connectors_by_name:
                    self._gateway.available_connectors_by_name[connector_name].close()
                self._gateway.load_connectors(self._get_general_config_in_local_format())
                self._gateway.connect_with_connectors()
            else:
                if 'keepaliveTimeMs' in self.grpc_configuration:
                    self.grpc_configuration['keepAliveTimeMs'] = self.grpc_configuration.pop('keepaliveTimeMs')
                if 'keepaliveTimeoutMs' in self.grpc_configuration:
                    self.grpc_configuration['keepAliveTimeoutMs'] = self.grpc_configuration.pop('keepaliveTimeoutMs')
                if 'keepalivePermitWithoutCalls' in self.grpc_configuration:
                    self.grpc_configuration['keepAlivePermitWithoutCalls'] = self.grpc_configuration.pop(
                        'keepalivePermitWithoutCalls')
                self.grpc_configuration = config
                with open(self._gateway.get_config_path() + "tb_gateway.json", "w", encoding="UTF-8") as file:
                    file.writelines(dumps(self._get_general_config_in_local_format(), indent='  '))
                self._gateway.send_attributes({'grpc_configuration': self.grpc_configuration})

                self.__log.debug('Processed GRPC configuration update successfully')

    def _handle_logs_configuration_update(self, config):
        self.__log.debug('Processing logs configuration update...')
        try:
            TbLogger.check_and_update_file_handlers_class_name(config)
            self.__log = getLogger('service')
            logs_conf_file_path = self._gateway.get_config_path() + 'logs.json'
            target_handlers = {}
            original_in_config = deepcopy(config)

            for handler in config['handlers']:
                filename = config['handlers'][handler].get('filename', None)

                if "consoleHandler" == handler or (filename is not None and os.path.exists(filename)):
                    target_handlers[handler] = config['handlers'][handler]
                elif filename is not None:
                    self.__log.debug('Handler %s not found. Trying to create...', handler)
                    try:
                        with open(filename, 'w'):
                            pass

                        target_handlers[handler] = config['handlers'][handler]
                    except Exception as e:
                        self.__log.error('Cannot create handler %s with %s error. Skipping...', handler, e)
                        self._delete_handler(config, handler)

                    self.__log.debug('Handler %s created.', handler)
                else:
                    self.__log.warning('Config is invalid. Filename is empty in handler %s', handler)
                    self._delete_handler(config, handler)

            config['handlers'] = target_handlers

            for logger in config['loggers']:

                if config['loggers'][logger].get('level', "").lower() in ('none', 'notset'):
                    config['loggers'][logger]['level'] = 100

            dictConfig(config)

            with open(logs_conf_file_path, 'w') as logs:
                logs.write(dumps(config, indent='  '))

            self._gateway.update_loggers()

            self.__log.debug("Logs configuration has been updated.")
            self._gateway.send_attributes({'logs_configuration': original_in_config})
        except Exception as e:
            self.__log.error("Remote logging configuration is wrong, cannot apply it!")
            self.__log.exception(e)

    def _handle_active_connectors_update(self, config):
        self.__log.debug('Processing active connectors configuration update...')

        for connector_name in config:
            self._gateway._check_shared_attributes(shared_keys=[connector_name])

        has_changed = False
        for_deletion = []
        for_deletion_ids = []
        for active_connector_name in self._gateway.available_connectors_by_name:
            if active_connector_name not in config:
                try:
                    connector_to_close = self._gateway.available_connectors_by_name[active_connector_name]
                    connector_to_close.close()
                    for device in self._gateway.get_connector_devices(connector_to_close):
                        self._gateway.del_device(device, False)
                    connector_id = connector_to_close.get_id()
                    self._gateway.available_connectors_by_id.pop(connector_id, None)
                    for_deletion.append(active_connector_name)
                    for_deletion_ids.append(connector_id)
                    if self._gateway._report_strategy_service is not None:
                        if self._gateway.available_connectors_by_name.get(active_connector_name) is not None and \
                            connector_id != self._gateway.available_connectors_by_name[active_connector_name].get_id():
                            another_connector_id = self._gateway.available_connectors_by_name[active_connector_name].get_id()
                            self._gateway._report_strategy_service.delete_all_records_for_connector_by_connector_id_and_connector_name(another_connector_id, active_connector_name)  # noqa
                        self._gateway._report_strategy_service.delete_all_records_for_connector_by_connector_id_and_connector_name(connector_id, active_connector_name)  # noqa
                    has_changed = True
                except Exception as e:
                    self.__log.exception("Exception on removing connector occurred:", exc_info=e)


        if has_changed:
            for name in for_deletion:
                self._gateway.available_connectors_by_name.pop(name, None)
            for connector_id in for_deletion_ids:
                self._gateway.available_connectors_by_id.pop(connector_id, None)

            self._delete_connectors_from_config(config)
            with open(self._gateway.get_config_path() + 'tb_gateway.json', 'w') as file:
                file.writelines(dumps(self._get_general_config_in_local_format(), indent='  '))
            self._active_connectors = config

        self._gateway.send_attributes({'active_connectors': config})

    def _handle_connector_configuration_update(self, config):
        if not isinstance(config, dict):
            return
        """
        Expected the following data structure:
        {
            "name": "Mqtt Broker Connector",
            "id": "bee7caf1-fd37-4026-8782-d480915d4d1a"
            "type": "mqtt",
            "configuration": "mqtt.json",
            "enableRemoteLogging": false,
            "logLevel": "INFO",
            "key?type=>grpc": "auto",
            "class?type=>custom": "",
            "configurationJson": {
                ...
            }
        }
        """

        self.__log.debug('Processing connectors configuration update...')

        try:
            connector_name = config['name']
            connector_type = config['type']
            if config.get('configuration') is None:
                config_file_name = connector_name.replace(' ', '_').lower() + '.json'
                config['configuration'] = config_file_name
            else:
                config_file_name = config['configuration']

            primary_identifier_parameter = 'id' if config.get('configurationJson', {}).get('id') else 'name'
            secondary_identifier_parameter = 'name' if primary_identifier_parameter == 'id' else None
            found_connectors = []
            for connector in self.connectors_configuration:
                connector_identifier = connector.get(primary_identifier_parameter)
                if connector_identifier is None:
                    self.__log.warning('Connector %s has no identifier parameter %s, it will be skipped.',
                                       connector,
                                       primary_identifier_parameter)
                    continue
                if (connector[primary_identifier_parameter] == config.get('configurationJson', {})
                        .get(primary_identifier_parameter)
                        or connector[primary_identifier_parameter] == config.get(primary_identifier_parameter)):
                    self.__log.debug('Connector %s found by primary identifier %s, %s',
                                     connector, primary_identifier_parameter, config.get(primary_identifier_parameter))
                    found_connectors.append(connector)
                elif secondary_identifier_parameter is not None \
                        and connector[secondary_identifier_parameter] == config.get(secondary_identifier_parameter):
                    self.__log.debug('Connector %s found by secondary identifier %s, %s',
                                     connector, secondary_identifier_parameter,
                                     config.get(secondary_identifier_parameter))
                    found_connectors.append(connector)

            if (config.get('configurationJson', {})
                    and config.get('configurationJson', {}).get('id') is None
                    and len(found_connectors) > 0
                    and found_connectors[0].get('configurationJson') is not None
                    and found_connectors[0].get('configurationJson', {}).get('id') is not None):
                connector_id = TBUtility.get_or_create_connector_id(found_connectors[0].get("configurationJson"))
            else:
                connector_id = TBUtility.get_or_create_connector_id(config.get('configurationJson'))

            if not found_connectors:
                connector_configuration = {'name': connector_name,
                                           'type': connector_type,
                                           'id': connector_id,
                                           'enableRemoteLogging': config.get('enableRemoteLogging', False),
                                           'configuration': config_file_name,
                                           CONFIG_VERSION_PARAMETER: config.get(CONFIG_VERSION_PARAMETER)}

                if config.get(REPORT_STRATEGY_PARAMETER):
                    connector_configuration[REPORT_STRATEGY_PARAMETER] = config[REPORT_STRATEGY_PARAMETER]

                if config.get('key'):
                    connector_configuration['key'] = config['key']

                if config.get('class'):
                    connector_configuration['class'] = config['class']

                locally_formatted_connector_config = {'logLevel': config['logLevel'],
                                                      'name': connector_name,
                                                      'enableRemoteLogging': config.get('enableRemoteLogging', False),
                                                      'id': connector_id}
                if config.get(REPORT_STRATEGY_PARAMETER) is not None:
                    locally_formatted_connector_config[REPORT_STRATEGY_PARAMETER] = config[REPORT_STRATEGY_PARAMETER]

                if config.get(CONFIG_VERSION_PARAMETER) is not None:
                    locally_formatted_connector_config[CONFIG_VERSION_PARAMETER] = config[CONFIG_VERSION_PARAMETER]

                with open(self._gateway.get_config_path() + config_file_name, 'w') as file:
                    config['configurationJson'].update(locally_formatted_connector_config)
                    self.create_configuration_file_backup(config, config_file_name)
                    file.writelines(dumps(config['configurationJson'], indent='  '))

                self.connectors_configuration.append(connector_configuration)
                with open(self._gateway.get_config_path() + 'tb_gateway.json', 'w') as file:
                    file.writelines(dumps(self._get_general_config_in_local_format(), indent='  '))

                self._gateway.load_connectors(self._get_general_config_in_local_format())

                if self._gateway._report_strategy_service is not None:
                    if self._gateway.available_connectors_by_name.get(connector_name) is not None and \
                            connector_id != self._gateway.available_connectors_by_name[connector_name].get_id():
                        another_connector_id = self._gateway.available_connectors_by_name[connector_name].get_id()
                        self._gateway._report_strategy_service.delete_all_records_for_connector_by_connector_id_and_connector_name(
                            another_connector_id, connector_name)  # noqa
                    self._gateway._report_strategy_service.delete_all_records_for_connector_by_connector_id_and_connector_name(
                        connector_id, connector_name)  # noqa

                self._gateway.connect_with_connectors()
            else:
                found_connector = found_connectors[0]
                changed = False

                config_file_path = self._gateway.get_config_path() + config_file_name
                if os.path.exists(config_file_path):
                    with open(config_file_path, 'r') as file:
                        connector_config_data = load(file)
                        config_hash = hash(str(connector_config_data))

                    if config_hash != hash(str(config['configurationJson'])):
                        self.create_configuration_file_backup(connector_config_data, config_file_name)
                        changed = True

                connector_configuration = None
                if (found_connector.get('id') != connector_id
                        or found_connector.get('name') != connector_name
                        or found_connector.get('type') != connector_type
                        or found_connector.get('class') != config.get('class')
                        or found_connector.get('key') != config.get('key')
                        or found_connector.get('configurationJson', {}).get('logLevel') != config.get('logLevel')
                        or found_connector.get('enableRemoteLogging', False) != config.get('enableRemoteLogging',
                                                                                           False)):
                    changed = True
                    connector_configuration = {'name': connector_name,
                                               'type': connector_type,
                                               'id': connector_id,
                                               'enableRemoteLogging': config.get('enableRemoteLogging', False),
                                               'configuration': config_file_name,
                                               CONFIG_VERSION_PARAMETER: config.get(CONFIG_VERSION_PARAMETER),
                                               }
                    if config.get(REPORT_STRATEGY_PARAMETER):
                        connector_configuration[REPORT_STRATEGY_PARAMETER] = config[REPORT_STRATEGY_PARAMETER]

                    if config.get('key'):
                        connector_configuration['key'] = config['key']

                    if config.get('class'):
                        connector_configuration['class'] = config['class']

                    found_connector.update(connector_configuration)

                if changed:
                    with open(self._gateway.get_config_path() + config_file_name, 'w') as file:
                        config_json_update = {
                            'logLevel': config['logLevel'],
                            'name': connector_name,
                            'enableRemoteLogging': config.get('enableRemoteLogging', False),
                            'id': connector_id,
                            CONFIG_VERSION_PARAMETER: config.get(CONFIG_VERSION_PARAMETER)
                            }
                        if config.get(REPORT_STRATEGY_PARAMETER):
                            config_json_update[REPORT_STRATEGY_PARAMETER] = config[REPORT_STRATEGY_PARAMETER]
                        config['configurationJson'].update(config_json_update)
                        file.writelines(dumps(config['configurationJson'], indent='  '))

                    if connector_configuration is None:
                        connector_configuration = found_connector

                    if connector_configuration.get('id') in self._gateway.available_connectors_by_id:
                        try:
                            retrieved_connector = self._gateway.available_connectors_by_id[connector_configuration['id']]
                            close_start = monotonic()
                            while not retrieved_connector.is_stopped(): # noqa
                                retrieved_connector.close()
                                if self._gateway.tb_client.is_connected():
                                    for device in self._gateway.get_connector_devices(retrieved_connector):
                                        self._gateway.del_device(device, False)
                                self._gateway.clean_shared_attributes_cache_for_connector_devices(retrieved_connector)
                                if monotonic() - close_start > 5:
                                    self.__log.error('Connector %s not stopped in 5 seconds', connector_configuration['id']) # noqa
                                    break
                        except Exception as e:
                            self.__log.exception("Exception on closing connector occurred:", exc_info=e)
                    elif connector_configuration.get('name') in self._gateway.available_connectors_by_name:
                        try:
                            retrieved_connector = self._gateway.available_connectors_by_name[connector_configuration['name']]
                            close_start = monotonic()
                            while not retrieved_connector.is_stopped():
                                retrieved_connector.close()
                                if self._gateway.tb_client.is_connected():
                                    for device in self._gateway.get_connector_devices(retrieved_connector):
                                        self._gateway.del_device(device, False)
                                self._gateway.clean_shared_attributes_cache_for_connector_devices(retrieved_connector)
                                if monotonic() - close_start > 5:
                                    self.__log.error('Connector %s not stopped in 5 seconds',
                                                     connector_configuration['name'])
                                    break
                        except Exception as e:
                            self.__log.exception("Exception on closing connector occurred:", exc_info=e)
                    else:
                        self.__log.warning('Connector with id %s not found in available connectors',
                                           connector_configuration.get('id'))
                    if connector_configuration.get('id') in self._gateway.available_connectors_by_id:
                        self._gateway.available_connectors_by_id.pop(connector_configuration['id'])
                        connector_configuration['id'] = connector_id
                    elif connector_configuration.get('name') in self._gateway.available_connectors_by_name:
                        self._gateway.available_connectors_by_name.pop(connector_configuration['name'])
                        connector_configuration['id'] = connector_id
                    else:
                        self.__log.warning('Connector with id %s not found in available connectors',
                                           connector_configuration.get('id'))

                    self._gateway.load_connectors(self._get_general_config_in_local_format())

                    if self._gateway._report_strategy_service is not None:
                        if self._gateway.available_connectors_by_name.get(connector_name) is not None and \
                            connector_id != self._gateway.available_connectors_by_name[connector_name].get_id():
                            another_connector_id = self._gateway.available_connectors_by_name[connector_name].get_id()
                            self._gateway._report_strategy_service.delete_all_records_for_connector_by_connector_id_and_connector_name(another_connector_id, connector_name)  # noqa
                        self._gateway._report_strategy_service.delete_all_records_for_connector_by_connector_id_and_connector_name(connector_id, connector_name)  # noqa

                    self._gateway.connect_with_connectors()

            # can be removed in the future versions:
            # config['sendDataOnlyOnChange'] = config['configurationJson'].get('sendDataOnlyOnChange', False)

            if connector_type != 'modbus':
                config['configurationJson'].pop('reportStrategy', None)

            for (device_name, device_config) in list(self._gateway.get_devices().items()):
                if (connector_id == device_config.get('connector').get_id()
                        and self._gateway.available_connectors_by_id.get(connector_id) is not None):
                    self._gateway.update_device(device_name,
                                                "connector",
                                                self._gateway.available_connectors_by_id[connector_id])

            self._gateway.send_attributes({connector_name: config})
            with open(self._gateway.get_config_path() + 'tb_gateway.json', 'w') as file:
                file.writelines(dumps(self._get_general_config_in_local_format(), indent='  '))
        except Exception as e:
            self.__log.exception("Exception on connector configuration update occurred:", exc_info=e)

    def _handle_remote_logging_level_update(self, config):
        self._gateway.send_attributes({'RemoteLoggingLevel': config})

    # HANDLERS SUPPORT METHODS -----------------------------------------------------------------------------------------
    def _apply_connection_config(self, config) -> bool:
        old_tb_client_config_path = self._gateway.tb_client.get_config_folder_path()
        old_tb_client_config = self._gateway.tb_client.config
        connection_logger = getLogger('tb_connection')
        try:
            self._gateway.tb_client.disconnect()
            self._gateway.tb_client.stop()
            existing_tls_context = self._gateway.tb_client.client._client._ssl_context

            while not self._gateway.stopped and self._gateway.tb_client.client.is_connected():
                sleep(1)

            apply_start = time()

            connection_state = False
            use_new_config = True

            config['rateLimits'] = old_tb_client_config.get('rateLimits', 'DEFAULT_TELEMETRY_RATE_LIMIT')
            config['dpRateLimits'] = old_tb_client_config.get('dpRateLimits', 'DEFAULT_TELEMETRY_DP_RATE_LIMIT')
            if 'messages_rate_limit' in inspect.signature(TBGatewayMqttClient.__init__).parameters:
                config['messagesRateLimits'] = old_tb_client_config.get('messagesRateLimits',
                                                                        'DEFAULT_MESSAGES_RATE_LIMIT')
                config['deviceMessagesRateLimits'] = old_tb_client_config.get('deviceMessagesRateLimits',
                                                                              'DEFAULT_MESSAGES_RATE_LIMIT')
                config['deviceRateLimits'] = old_tb_client_config.get('deviceRateLimits',
                                                                      'DEFAULT_TELEMETRY_RATE_LIMIT')
                config['deviceDpRateLimits'] = old_tb_client_config.get('deviceDpRateLimits',
                                                                        'DEFAULT_TELEMETRY_DP_RATE_LIMIT')

            previous_rate_limits = self._gateway.tb_client.get_rate_limits()

            # check if security type is tlsAccessToken
            config = self.__check_and_process_tls_access_token(config)

            while not self._gateway.stopped and not connection_state:
                self._gateway.__subscribed_to_rpc_topics = False
                if use_new_config:
                    new_tb_client = TBClient(config, old_tb_client_config_path, connection_logger)
                else:
                    new_tb_client = TBClient(old_tb_client_config, old_tb_client_config_path, connection_logger)
                new_tb_client.update_client_rate_limits_with_previous(previous_rate_limits)
                new_tb_client.client._client._ssl_context = existing_tls_context
                new_tb_client.connect(timeout=5)
                while not self._gateway.stopped and time() - apply_start <= 30 and not connection_state:
                    connection_state = new_tb_client.is_connected()
                    sleep(.1)

                if connection_state:
                    self._gateway.tb_client = new_tb_client
                    self._gateway.__subscribed_to_rpc_topics = False
                    self._gateway.subscribe_to_required_topics()

                    return True
                else:
                    new_tb_client.disconnect()
                    new_tb_client.stop()
                    while not self._gateway.stopped and new_tb_client.client.is_connected():
                        sleep(1)
                    apply_start = time() * 1000
                    use_new_config = not use_new_config
                if self._gateway.stopped:
                    return False
        except Exception as e:
            self.__log.exception(e)
            self._revert_connection()
            return False

    def __check_and_process_tls_access_token(self, config):
        if config.get('security', {}).get('type') == 'tlsAccessToken':
            cert_content = config['security']['caCert']

            ca_cert_path = self._gateway.get_config_path() + 'ca.pem'
            with open(ca_cert_path, 'w') as file:
                file.write(cert_content)

            config['security']['caCert'] = ca_cert_path

        return config

    def _revert_connection(self):
        try:
            self.__log.warning("Remote general configuration will be restored.")
            self._gateway.tb_client.disconnect()
            self._gateway.tb_client.stop()
            connection_logger = getLogger('tb_connection')
            self._gateway.tb_client = TBClient(self.general_configuration,
                                               self._gateway.get_config_path(),
                                               connection_logger)
            self._gateway.tb_client.connect()
            self._gateway.subscribe_to_required_topics()
            self.__log.debug("%s connection has been restored", str(self._gateway.tb_client.client))
        except Exception as e:
            self.__log.exception("Exception on reverting configuration occurred:")
            self.__log.exception(e)

    def _apply_statistics_config(self, config) -> bool:
        try:
            commands = config.get('commands', [])
            if commands:
                statistics_conf_file_name = self.general_configuration['statistics'].get('configuration',
                                                                                         'statistics.json')
                if statistics_conf_file_name is None:
                    statistics_conf_file_name = 'statistics.json'

                with open(self._gateway.get_config_path() + statistics_conf_file_name, 'w') as file:
                    file.writelines(dumps(commands, indent='  '))
                config['configuration'] = statistics_conf_file_name

            self._gateway.init_statistics_service(config)
            self.general_configuration['statistics'] = config
            return True
        except Exception as e:
            self.__log.error('Something went wrong with applying the new statistics configuration. Reverting...')
            self.__log.exception(e)
            self._gateway.init_statistics_service(
                self.general_configuration.get('statistics', {'enable': True, 'statsSendPeriodInSeconds': 3600}))
            return False

    def _apply_device_filtering_config(self, config):
        try:
            self._gateway.init_device_filtering(config.get('deviceFiltering', {'enable': False}))
            return True
        except Exception as e:
            self.__log.error('Something went wrong with applying the new device filtering configuration. Reverting...')
            self.__log.exception(e)
            self._gateway.init_device_filtering(
                self.general_configuration.get('deviceFiltering', {'enable': False}))
            return False

    def _apply_remote_shell_config(self, config):
        try:
            self._gateway.init_remote_shell(config.get('remoteShell'))
            return True
        except Exception as e:
            self.__log.error('Something went wrong with applying the new Remote Shell configuration. Reverting...')
            self.__log.exception(e)
            self._gateway.init_remote_shell(self.general_configuration.get('remoteShell'))
            return False

    def _apply_report_strategy_config(self, config):
        old_main_report_strategy = None
        if self._gateway.get_report_strategy_service() is not None:
            old_main_report_strategy = self._gateway.get_report_strategy_service().main_report_strategy
        try:
            new_main_report_strategy = ReportStrategyConfig(config.get(REPORT_STRATEGY_PARAMETER,
                                                                       DEFAULT_REPORT_STRATEGY_CONFIG))
            if ReportStrategy.DISABLED != new_main_report_strategy.report_strategy:
                if self._gateway.get_report_strategy_service() is not None:
                    self._gateway.get_report_strategy_service().stop_event.set()
                    self._gateway.get_report_strategy_service().clear_cache()
                self._gateway._report_strategy_service = ReportStrategyService(config,
                                                                               self._gateway,
                                                                               self._gateway.get_converted_data_queue(),
                                                                               self.__log)
                self._gateway.get_report_strategy_service().main_report_strategy = new_main_report_strategy
                self._gateway.get_report_strategy_service().clear_cache()
                for connector in self._gateway.available_connectors_by_id.values():
                    try:
                        connector_report_strategy = ReportStrategyConfig(connector.get_config().get(REPORT_STRATEGY_PARAMETER, {}))
                        self._gateway.get_report_strategy_service().register_connector_report_strategy(connector_name=connector.get_name(),
                                                                                                       connector_id=connector.get_id(),
                                                                                                       report_strategy_config=connector_report_strategy)
                    except ValueError:
                        pass
            else:
                self._gateway.get_report_strategy_service().stop_event.set()
                self._gateway.get_report_strategy_service().clear_cache()
                self._gateway._report_strategy_service = None
            return True
        except Exception as e:
            if old_main_report_strategy is not None:
                self._gateway.get_report_strategy_service().main_report_strategy = old_main_report_strategy
            self.__log.error('Something went wrong with applying the new Report Strategy configuration. Reverting...')
            self.__log.exception(e)
            return False

    def _apply_other_params_config(self, config):
        self._gateway.config['thingsboard'].update(config)

    def _delete_connectors_from_config(self, connector_list):
        self._config['connectors'] = list(
            filter(lambda connector: connector['name'] in connector_list, self.connectors_configuration))

    def _check_statistics_configuration_changes(self, config):
        general_statistics_config = self.general_configuration.get('statistics', self.DEFAULT_STATISTICS)
        for key in config.keys():
            if (key not in ('commands', 'configuration')
                    and (key not in general_statistics_config or config[key] != general_statistics_config[key])):
                return True
        for key in general_statistics_config.keys():
            if key not in config:
                return True

        commands = []
        if general_statistics_config.get('configuration'):
            with open(self._gateway.get_config_path() + general_statistics_config['configuration'], 'r') as file:
                commands = load(file)

        if config.get('commands', []) != commands:
            return True

        return False

    def _cleanup(self):
        self.general_configuration['statistics'].pop('commands')

    def _is_modified(self, attr_name, config):
        try:
            file_path = config.get('configuration') or self._modifiable_static_attrs.get(attr_name)
        except AttributeError:
            file_path = None

        # if there is no file path that means that it is RemoteLoggingLevel or active_connectors attribute update
        # in this case, we have to update the configuration without TS compare
        if file_path is None:
            return True

        try:
            file_path = self._gateway.get_config_path() + file_path
            if config.get('ts', 0) <= int(os.path.getmtime(file_path) * 1000):
                return False
        except OSError:
            self.__log.warning('File %s not exist', file_path)

        return True

    def _is_running(self, attr_name, config):
        if (config.get(attr_name, {}).get('configurationJson', {}).get('id') in self._gateway.available_connectors_by_id
                or attr_name in self._gateway.available_connectors_by_name):
            return True
        return False

    def create_configuration_file_backup(self, config_data, config_file_name):
        backup_folder_path = self._gateway.get_config_path() + "backup"
        if not os.path.exists(backup_folder_path):
            os.mkdir(backup_folder_path)

        backup_file_name = config_file_name.split('.')[0] + ".backup." + str(int(time())) + ".json"
        backup_file_path = backup_folder_path + os.path.sep + backup_file_name
        with open(backup_file_path, "w") as backup_file:
            self.__log.debug(f"Backup file created for configuration file {config_file_name} in {backup_file_path}")
            backup_file.writelines(dumps(config_data, indent='  ', skipkeys=True))
        self.check_and_remove_old_backups_for_configuration_file(backup_folder_path, config_file_name)

    def _create_connectors_backup(self):
        for connector in self.connectors_configuration:
            if connector.get('configurationJson'):
                self.create_configuration_file_backup(connector['configurationJson'], connector['configuration'])
            else:
                self.__log.debug(f"Configuration for {connector['name']} connector is not found, backup wasn't created")

    @staticmethod
    def check_and_remove_old_backups_for_configuration_file(backup_folder_path, config_file_name):
        backup_files = [f for f in os.listdir(backup_folder_path) if f.startswith(config_file_name.split('.')[0]) and f.endswith('.json')] # noqa
        if len(backup_files) > RemoteConfigurator.ALLOWED_BACKUPS_AMOUNT:
            backup_files.sort()
            for i in range(len(backup_files) - RemoteConfigurator.ALLOWED_BACKUPS_AMOUNT):
                os.remove(backup_folder_path + os.path.sep + backup_files[i])

    @staticmethod
    def _delete_handler(config, handler):
        for logger in config['loggers']:
            if handler in config['loggers'][logger]['handlers']:
                config['loggers'][logger]['handlers'].remove(handler)

    def __is_running(self, request_config):
        return (request_config.get('configurationJson', {}).get('id') in self._gateway.available_connectors_by_id or
                request_config.get('name') in self._gateway.available_connectors_by_name)

    @staticmethod
    def __is_security_match(old_security, new_security):
        if new_security.get('type'):
            is_match = RemoteConfigurator.__is_security_match_by_type(old_security, new_security)
        else:
            is_match = RemoteConfigurator.__is_security_match_by_fields(old_security, new_security)

        return is_match

    @staticmethod
    def __is_security_match_by_type(old_security, new_security):
        new_security_type = new_security.get('type')

        if new_security_type == 'accessToken':
            return RemoteConfigurator.__is_access_token_match(old_security.get('accessToken'),
                                                              new_security.get('accessToken'))
        elif new_security_type == 'usernamePassword':
            return RemoteConfigurator.__is_username_password_match(old_security, new_security)
        elif new_security_type == 'tlsAccessToken':
            return RemoteConfigurator.__is_ca_cert_match(old_security, new_security)
        else:
            return False

    @staticmethod
    def __is_security_match_by_fields(old_security, new_security):
        if new_security.get('accessToken'):
            return RemoteConfigurator.__is_access_token_match(old_security.get('accessToken'),
                                                              new_security.get('accessToken'))
        elif new_security.get('username'):
            return RemoteConfigurator.__is_username_password_match(old_security, new_security)
        elif new_security.get('caCert'):
            return RemoteConfigurator.__is_ca_cert_match(old_security, new_security)
        else:
            return False

    @staticmethod
    def __is_access_token_match(old_access_token, new_access_token):
        return old_access_token == new_access_token

    @staticmethod
    def __is_username_password_match(old_security, new_security):
        return new_security.get('username') == old_security.get('username') and \
               new_security.get('password') == old_security.get('password') and \
               new_security.get('clientId') == old_security.get('clientId')

    @staticmethod
    def __is_ca_cert_match(old_security, new_security):
        return new_security.get('accessToken') == old_security.get('accessToken') and \
            new_security.get('caCert') == old_security.get('caCert')



#      Copyright 2025. ThingsBoard
#  #
#      Licensed under the Apache License, Version 2.0 (the "License");
#      you may not use this file except in compliance with the License.
#      You may obtain a copy of the License at
#  #
#          http://www.apache.org/licenses/LICENSE-2.0
#  #
#      Unless required by applicable law or agreed to in writing, software
#      distributed under the License is distributed on an "AS IS" BASIS,
#      WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#      See the License for the specific language governing permissions and
#      limitations under the License.

from thingsboard_gateway.gateway.statistics.service_functions import StatisticsServiceFunctions

ONCE_SEND_STATISTICS_CONFIG = [
    {
        "function": StatisticsServiceFunctions.total_memory,
        "attributeOnGateway": "totalMemory"
    },
    {
        "function": StatisticsServiceFunctions.total_disk_memory,
        "attributeOnGateway": "totalDisk"
    },
    {
        "function": StatisticsServiceFunctions.total_cpu_count,
        "attributeOnGateway": "cpuCount"
    }
]

SERVICE_STATS_CONFIG = [
    {
        "function": StatisticsServiceFunctions.storage_msgs_pulled,
        "attributeOnGateway": "storageMsgPulled"
    },
    {
        "function": StatisticsServiceFunctions.storage_msgs_count,
        "attributeOnGateway": "storageMsgCount"
    },
    {
        "function": StatisticsServiceFunctions.platform_msgs_pushed,
        "attributeOnGateway": "platformMsgPushed"
    },
    {
        "function": StatisticsServiceFunctions.platform_attr_produced,
        "attributeOnGateway": "platformAttrProduced"
    },
    {
        "function": StatisticsServiceFunctions.platform_ts_produced,
        "attributeOnGateway": "platformTsProduced"
    }
]

MACHINE_STATS_CONFIG = [
    {
        "function": StatisticsServiceFunctions.cpu_usage,
        "attributeOnGateway": "totalCpuUsage"
    },
    {
        "function": StatisticsServiceFunctions.disk_usage_perc,
        "attributeOnGateway": "diskUsage"
    },
    {
        "function": StatisticsServiceFunctions.ram_usage,
        "attributeOnGateway": "freeMemory"
    },
    {
        "function": StatisticsServiceFunctions.free_disk_space,
        "attributeOnGateway": "freeDisk"
    },
    {
        "function": StatisticsServiceFunctions.gateway_cpu_usage,
        "attributeOnGateway": "gwProcessCpuUsage"
    },
    {
        "function": StatisticsServiceFunctions.gateway_process_memory_usage_percent,
        "attributeOnGateway": "gwMemory"
    },
    {
        "function": StatisticsServiceFunctions.gateway_process_memory_full_info,
        "attributeOnGateway": "gwProcessMemoryFullInfo"
    },
    {
        "function": StatisticsServiceFunctions.msgs_sent_to_platform,
        "attributeOnGateway": "msgsSentToPlatform"
    },
    {
        "function": StatisticsServiceFunctions.msgs_received_from_platform,
        "attributeOnGateway": "msgsReceivedFromPlatform"
    }
]



#     Copyright 2025. ThingsBoard
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

import asyncio
from asyncio import CancelledError
from threading import Thread
from time import monotonic, sleep

from pymodbus.datastore import ModbusSparseDataBlock, ModbusServerContext
from pymodbus.device import ModbusDeviceIdentification
from pymodbus.framer.base import FramerType
from pymodbus.server import (
    StartAsyncTcpServer,
    StartAsyncTlsServer,
    StartAsyncUdpServer,
    StartAsyncSerialServer,
    ServerAsyncStop
)
from pymodbus import __version__ as pymodbus_version

from thingsboard_gateway.connectors.modbus.bytes_modbus_downlink_converter import BytesModbusDownlinkConverter
from thingsboard_gateway.connectors.modbus.entities.bytes_downlink_converter_config import \
    BytesDownlinkConverterConfig
from thingsboard_gateway.connectors.modbus.constants import PymodbusDefaults
from thingsboard_gateway.connectors.modbus.entities.slave_context import SlaveContext
from thingsboard_gateway.connectors.modbus.constants import (
    ADDRESS_PARAMETER,
    BYTE_ORDER_PARAMETER,
    FUNCTION_CODE_SLAVE_INITIALIZATION,
    FUNCTION_TYPE,
    FUNCTION_CODE_READ,
    HOST_PARAMETER,
    IDENTITY_SECTION,
    METHOD_PARAMETER,
    OBJECTS_COUNT_PARAMETER,
    PORT_PARAMETER,
    REPACK_PARAMETER,
    SERIAL_CONNECTION_TYPE_PARAMETER,
    TAG_PARAMETER,
    WORD_ORDER_PARAMETER
)
from thingsboard_gateway.gateway.constants import DEVICE_NAME_PARAMETER, TYPE_PARAMETER

SLAVE_TYPE = {
    'tcp': StartAsyncTcpServer,
    'tls': StartAsyncTlsServer,
    'udp': StartAsyncUdpServer,
    'serial': StartAsyncSerialServer
}


class Server(Thread):
    def __init__(self, config, logger):
        super().__init__()
        self.daemon = True
        self.name = 'Gateway Modbus Server (Slave)'

        self.__log = logger

        self.__config = config

        self.device_name = config.get('deviceName', 'Modbus Slave')
        self.device_type = config.get('deviceType', 'default')
        self.poll_period = config.get('pollPeriod', 5000)

        self.__type = config.get('type', 'tcp').lower()
        self.__identity = self.__get_identity(self.__config)
        self.__server_context = self.__get_server_context(self.__config)
        self.__connection_config = self.__get_connection_config(self.__config)

        try:
            self.loop = asyncio.new_event_loop()
            asyncio.set_event_loop(self.loop)
        except RuntimeError:
            self.loop = asyncio.get_event_loop()

    def __str__(self):
        return self.name

    def run(self):
        try:
            self.loop.run_until_complete(self.start_server())
        except CancelledError:
            self.__log.debug("Server %s has been stopped", self.name)
        except Exception as e:
            self.__log.error("Server has been stopped with error: %s", e)

    def stop(self):
        asyncio.run_coroutine_threadsafe(self.__shutdown(), self.loop)

        self.__check_is_alive()

    def __check_is_alive(self):
        start_time = monotonic()

        while self.is_alive():
            if monotonic() - start_time > 10:
                self.__log.error("Failed to stop slave %s", self.name)
                break
            sleep(.1)

    async def __shutdown(self):
        await ServerAsyncStop()

    async def start_server(self):
        try:
            await SLAVE_TYPE[self.__type](identity=self.__identity, context=self.__server_context,
                                          **self.__connection_config)
        except Exception as e:
            self.__log.error('Failed to start Gateway Modbus Server (Slave): %s', e)

    def get_slave_config_format(self):
        """
        Function return configuration in slave format for adding it to gateway slaves list
        """

        config = {
            **self.__config,
            'deviceName': self.device_name,
            'deviceType': self.device_type,
            'pollPeriod': self.poll_period
        }

        for (register, register_values) in self.__config.get('values', {}).items():
            for (section_name, section_values) in register_values.items():
                if not config.get(section_name):
                    config[section_name] = []

                for item in section_values:
                    item_config = {
                        **item,
                        'functionCode': FUNCTION_CODE_READ[register]
                        if section_name not in ('attributeUpdates', 'rpc') else item['functionCode'],
                    }
                    config[section_name].append(item_config)

        return config

    @staticmethod
    def is_runnable(config):
        return config.get('slave') and config.get('slave', {}).get('sendDataToThingsBoard', False)

    @staticmethod
    def __get_identity(config):
        identity = None

        if config.get(IDENTITY_SECTION):
            identity = ModbusDeviceIdentification()
            identity.VendorName = config[IDENTITY_SECTION].get('vendorName', '')
            identity.ProductCode = config[IDENTITY_SECTION].get('productCode', '')
            identity.VendorUrl = config[IDENTITY_SECTION].get('vendorUrl', '')
            identity.ProductName = config[IDENTITY_SECTION].get('productName', '')
            identity.ModelName = config[IDENTITY_SECTION].get('ModelName', '')
            identity.MajorMinorRevision = pymodbus_version

        return identity

    @staticmethod
    def __get_connection_config(config):
        connection_config = {
            'framer': FramerType[config.get(METHOD_PARAMETER, 'socket').upper()],
        }

        if config[TYPE_PARAMETER] in ('tcp', 'udp'):
            connection_config[ADDRESS_PARAMETER] = (config.get(HOST_PARAMETER), int(config.get(PORT_PARAMETER)))
        elif config[TYPE_PARAMETER] == SERIAL_CONNECTION_TYPE_PARAMETER:
            connection_config[PORT_PARAMETER] = config.get(PORT_PARAMETER)
            connection_config['baudrate'] = int(config.get('baudrate', PymodbusDefaults.Baudrate))
            connection_config['bytesize'] = int(config.get('bytesize', PymodbusDefaults.Bytesize))
            connection_config['stopbits'] = int(config.get('stopbits', PymodbusDefaults.Stopbits))
            connection_config['parity'] = config.get('parity', PymodbusDefaults.Parity)
            connection_config['handle_local_echo'] = config.get('handleLocalEcho', False)

        if config.get('security'):
            security_config = config['security']
            connection_config['keyfile'] = security_config.get('keyfile')
            connection_config['certfile'] = security_config.get('certfile')
            connection_config['password'] = security_config.get('password')

        return connection_config

    def __get_server_context(self, config):
        blocks = {}
        if (config.get('values') is None) or (not len(config.get('values'))):
            self.__log.error("No values to read from device %s", config.get(DEVICE_NAME_PARAMETER, 'Modbus Slave'))
            return

        converter = BytesModbusDownlinkConverter({}, self.__log)
        for (key, value) in config.get('values').items():
            values = {}
            for section in ('attributes', 'timeseries', 'attributeUpdates', 'rpc'):
                for item in value.get(section, []):
                    try:
                        function_code = FUNCTION_CODE_SLAVE_INITIALIZATION[key][0] \
                            if item[OBJECTS_COUNT_PARAMETER] <= 1 else FUNCTION_CODE_SLAVE_INITIALIZATION[key][1]
                        converter_config = BytesDownlinkConverterConfig(
                            device_name=config.get(DEVICE_NAME_PARAMETER, 'Gateway'),
                            byte_order=config[BYTE_ORDER_PARAMETER],
                            word_order=config.get(WORD_ORDER_PARAMETER, 'LITTLE'),
                            repack=config.get(REPACK_PARAMETER, False),
                            objects_count=item[OBJECTS_COUNT_PARAMETER],
                            function_code=function_code,
                            lower_type=item.get(
                                TYPE_PARAMETER, item.get(TAG_PARAMETER, 'error')),
                            address=item.get(ADDRESS_PARAMETER, 0)
                        )
                        converted_value = converter.convert(
                            converter_config, {'data': {'params': item['value']}})
                        if converted_value is not None:
                            values[item[ADDRESS_PARAMETER] + 1] = converted_value
                        else:
                            self.__log.error("Failed to convert value %s with type %s, skipping...", item['value'],
                                             item['type'])
                    except Exception as e:
                        self.__log.error("Failed to configure value %s with error: %s, skipping...", item['value'], e)

                try:
                    if len(values):
                        blocks[FUNCTION_TYPE[key]] = ModbusSparseDataBlock(values)
                except Exception as e:
                    self.__log.error("Failed to configure block %s with error: %s", key, e)

        if not len(blocks):
            self.__log.info("%s - will be initialized without values",
                            config.get(DEVICE_NAME_PARAMETER, 'Modbus Slave'))

        return ModbusServerContext(slaves=SlaveContext(**blocks), single=True)



#     Copyright 2025. ThingsBoard
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.


from queue import Queue
from threading import Event, Thread, Lock
from typing import List, TYPE_CHECKING

import serial.tools
import serial.tools.list_ports
from thingsboard_gateway.tb_utility.tb_utility import TBUtility
from time import monotonic, sleep

try:
    import serial
except ImportError:
    print("pyserial library not found - installing...")
    TBUtility.install_package("pyserial")
    import serial

from thingsboard_gateway.connectors.connector import Connector
from thingsboard_gateway.tb_utility.tb_loader import TBModuleLoader
from thingsboard_gateway.tb_utility.tb_logger import init_logger

if TYPE_CHECKING:
    #  necessary for type checking to avoid circular import
    from thingsboard_gateway.gateway.tb_gateway_service import TBGatewayService


class SerialDevice(Thread):
    """
    Serial device class is used to represent a device that is connected to the serial port.
    It is used to read data from the device and send it to the platform.
    """
    def __init__(self, device_config, uplink_converter, stop_event: Event, logger, uplink_queue):
        super().__init__()
        self.__log = logger
        self.uplink_queue = uplink_queue
        self.daemon = True
        self.stopped = True
        self.__connector_stopped = stop_event
        self.config = device_config
        self.name = self.config.get('deviceName', self.config.get('name', 'SerialDevice'))
        self.type = self.config.get('deviceType', self.config.get('type', 'default'))
        self.uplink_converter = uplink_converter
        self.downlink_converter = None
        self.delimiter = self.config.get('delimiter', '\n')
        self.__rpc_in_progress = Event()
        self.__previous_connect = 0

        self.port = self.config.get('port', '/dev/ttyUSB0')
        self.baudrate = self.config.get('baudrate', 9600)
        self.timeout = self.config.get('timeout', 1)
        self.bytesize = self.config.get('bytesize', serial.EIGHTBITS)
        self.stopbits = self.config.get('stopbits', serial.STOPBITS_ONE)
        self.parity = self.config.get('parity', serial.PARITY_NONE)
        self.dsrdtr = self.config.get('dsrdtr', False)
        self.rtscts = self.config.get('rtscts', False)
        self.xonxoff = self.config.get('xonxoff', False)
        self.write_timeout = self.config.get('writeTimeout', None)
        self.inter_byte_timeout = self.config.get('interByteTimeout', None)
        self.exclusive = self.config.get('exclusive', None)
        self.__serial_lock = Lock()
        self.__serial = None

    def get_serial(self):
        """
        Method to get serial connection to the device.
        If connection is not established, it tries to connect to the device.
        """
        with self.__serial_lock:
            if self.__serial is None or not self.__serial.is_open:
                try:
                    self.__serial = serial.Serial(
                        port=self.port,
                        baudrate=self.baudrate,
                        timeout=self.timeout,
                        bytesize=self.bytesize,
                        stopbits=self.stopbits,
                        parity=self.parity,
                        dsrdtr=self.dsrdtr,
                        rtscts=self.rtscts,
                        xonxoff=self.xonxoff,
                        write_timeout=self.write_timeout,
                        inter_byte_timeout=self.inter_byte_timeout,
                        exclusive=self.exclusive
                    )
                    self.__log.info("Connected to device %s", self.name)
                except Exception as e:
                    self.__log.error("Failed to connect to device %s: %s", self.name, e)
                    self.__serial = None
        return self.__serial

    def run(self):
        """
        Main method to read data from the device and send it to the platform.
        """
        self.__log.info("Device %s started", self.name)
        self.stopped = False
        while not self.__connector_stopped.is_set() and not self.stopped:
            try:
                # Avoid excessive "Failed to connect to device" errors when __read_data_from_serial() calls get_serial()
                if self.get_serial() is None or not self.__serial.is_open:
                    sleep(3)
                else:
                    if not self.__rpc_in_progress.is_set():
                        data_from_device = self.__read_data_from_serial()
                        if data_from_device:
                            try:
                                converted_data = self.uplink_converter.convert(None, data_from_device)
                                self.uplink_queue.put(converted_data)
                            except Exception as e:
                                self.__log.error("Failed to convert data from device %s: %s", self.name, e)
            except Exception as e:
                self.__log.exception("Error in device %s: %s", self.name, e)
                self.stop()
        self.__log.info("Device %s stopped", self.name)

    def handle_rpc_request(self, rpc_method, params):
        """
        Method to process RPC requests from the platform.
        """
        result = {"success": True}
        processed = False
        for rpc_config in self.config.get("serverSideRpc", []):
            if rpc_method == rpc_config.get("method"):
                processed = True
                self.__rpc_in_progress.set()
                try:
                    if self.downlink_converter is not None:
                        converted_data = self.downlink_converter.convert(rpc_config, params)
                        if converted_data:
                            with_response = rpc_config.get("withResponse", False)
                            response_timeout = rpc_config.get("responseTimeoutSec", 5)
                            response = self.write(converted_data,
                                                  with_response=with_response,
                                                  response_timeout=response_timeout)
                            if with_response:
                                response_uplink_config = {}
                                if rpc_config.get("responseType"):
                                    response_uplink_config["type"] = rpc_config.get("responseType")
                                if rpc_config.get("responseFromByte"):
                                    response_uplink_config["fromByte"] = rpc_config.get("responseFromByte")
                                if rpc_config.get("responseToByte"):
                                    response_uplink_config["toByte"] = rpc_config.get("responseToByte")
                                if rpc_config.get("responseUntilDelimiter"):
                                    response_uplink_config["delimiter"] = rpc_config.get("responseUntilDelimiter")
                                if response_uplink_config and response:
                                    result = self.uplink_converter.convert(response_uplink_config, response)
                                else:
                                    result = {"error": "Cannot convert response with config: %r and response: %r" % (
                                        response_uplink_config, response), "success": False}
                        else:
                            result = {"error": "No data to send", "success": False}
                    else:
                        result = {"error": "Downlink converter not defined", "success": False}
                except Exception as e:
                    self.__log.error("Failed to process RPC with method: %r, params: %r, config: %r - Error: %s",
                                     rpc_method, params, rpc_config, e)
                    result = {"error": str(e), "success": False}
                finally:
                    self.__rpc_in_progress.clear()
        if not processed:
            result = {"error": "Method not found", "success": False}
        return result

    def write(self, data, with_response=False, response_timeout=5):
        """
        Method to write data to the device.
        If with_response is set to True, it waits for the response from the device.
        """
        try:
            serial_conn = self.get_serial()
            if serial_conn:
                with self.__serial_lock:
                    serial_conn.write(data)
                    self.__log.debug("Written to device %s: %s", self.name, data)
                if with_response:
                    return self.__read_data_from_serial(response_timeout)
        except Exception as e:
            self.__log.exception("Failed to write to device %s: %s", self.name, e)
        return None

    def __read_data_from_serial(self, timeout=1):
        """
        Method to read data from the device.
        It reads data until the delimiter is found.
        """
        data_from_device = b''
        serial_conn = None
        try:
            serial_conn = self.get_serial()
            if serial_conn and serial_conn.is_open: 
                # Ensure serial_conn is not None before accessing its timeout attribute
                previous_timeout = serial_conn.timeout
                while not data_from_device.endswith(self.delimiter.encode('utf-8')):
                    serial_conn.timeout = timeout
                    chunk = serial_conn.read(1)
                    if chunk:
                        data_from_device += chunk
                    if self.__connector_stopped.is_set() or not chunk or self.stopped:
                        break
        except Exception as e:
            self.__log.exception("Failed to read from device %s: %s", self.name, e)
        finally:
            if serial_conn:
                serial_conn.timeout = previous_timeout
        return data_from_device

    def stop(self):
        self.stopped = True
        with self.__serial_lock:
            if self.__serial:
                self.__serial.close()
                self.__serial = None

    def is_connected_reconnect_if_needed(self):
        """
        Method to check if the device is connected.
        If the device is not connected, it tries to reconnect.
        """
        if self.__serial is None or not self.__serial.is_open:
            if monotonic() - self.__previous_connect > 1:
                self.__previous_connect = monotonic()
                self.__log.info("Reconnecting to device %s", self.name)
                self.get_serial()
                # Only return True if the serial connection is successful
                return self.__serial and self.__serial.is_open
        else:
            return True


class SerialConnector(Thread, Connector):
    """
    Serial connector class is used to represent a serial connector.
    It is used to manage devices connected to the serial ports.
    """
    def __init__(self, gateway: 'TBGatewayService', config, connector_type):
        super().__init__()
        self._connector_type = connector_type  # required to have for get connector type method
        self.__config = config  # required to have for get config method
        self.__id = self.__config["id"]  # required to have for get id method
        self.__gateway = gateway  # required to have for send data to storage method or to use other gateway methods
        self.name = self.__config["name"]  # required to have for get name method
        self.__connected = False  # required to have for is connected method
        self.__uplink_queue = Queue(self.__config.get('uplinkQueueSize', 100000))
        self._log = init_logger(self.__gateway, self.name, level=self.__config.get('logLevel'),
                                enable_remote_logging=self.__config.get('enableRemoteLogging', False),
                                is_connector_logger=True)
        self._converter_log = init_logger(self.__gateway, self.name, level=self.__config.get('logLevel'),
                                          enable_remote_logging=self.__config.get('enableRemoteLogging', False),
                                          is_converter_logger=True)
        self._log.info("Starting %s connector", self.get_name())
        self.daemon = True
        self.stopped = Event()
        self.stopped.set()
        self.__devices: List[SerialDevice] = []
        self._log.info('Connector %s initialization success.', self.get_name())

    def __start_devices(self):
        failed_to_connect_devices = len(self.__devices)
        for device in self.__devices:
            try:
                device.start()
                failed_to_connect_devices -= 1
            except Exception as e:
                self._log.exception("Failed to start device %s, error: %s", device.name, e)
        self.__connected = failed_to_connect_devices == 0

    def open(self):
        """
        Service method to start the connector.
        """
        self.stopped.clear()
        self.start()

    def get_name(self):
        return self.name

    def get_type(self):
        return self._connector_type

    def is_connected(self):
        return self.__connected

    def is_stopped(self):
        return self.stopped.is_set()

    def get_config(self):
        return self.__config

    def get_id(self):
        return self.__id

    def __load_devices(self):
        """
        Method to create devices objects using configuration file and create converters for them.
        """
        devices_config = self.__config.get('devices')
        try:
            if devices_config is not None:
                for device_config in devices_config:
                    device = None
                    uplink_converter_class_name = device_config.get('converter', device_config.get('uplink_converter'))
                    if uplink_converter_class_name is not None:
                        converter_class = TBModuleLoader.import_module(self._connector_type,
                                                                       uplink_converter_class_name)
                        uplink_converter = converter_class(device_config, self._log)
                        device = SerialDevice(device_config, uplink_converter, self.stopped,
                                              self._log, self.__uplink_queue)
                    else:
                        self._log.error('Converter configuration for the connector %s -- \
                            not found, please check your configuration file.', self.get_name())
                    if device_config.get('downlink_converter') is not None:
                        downlink_converter_class = TBModuleLoader.import_module(self._connector_type,
                                                                                device_config.get('downlink_converter'))
                        if device is not None:
                            device.downlink_converter = downlink_converter_class(device_config, self._converter_log)
                    if device is not None:
                        self.__devices.append(device)
            else:
                self._log.error('Section "devices" in the configuration not found. \
                    A connector %s has being stopped.', self.get_name())
                self.close()
        except Exception as e:
            self._log.error('Failed to load devices, error: %s', e)

    def run(self):
        """
        Main method to manage devices connected to the serial ports and process data from them.
        """
        try:
            self.__load_devices()
            self.__start_devices()
            self._log.info("Devices in configuration file found: %s ",
                           '\n'.join(device.name for device in self.__devices))
            while not self.stopped.is_set():
                try:
                    connected_devices = len(self.__devices)
                    for device in self.__devices:
                        if not device.stopped and not device.is_connected_reconnect_if_needed():
                            connected_devices -= 1
                            self._log.error("Device %s is not connected", device.name)
                            device.stop()
                            device.join()
                            device = SerialDevice(device.config, device.uplink_converter, self.stopped,
                                                  self._log, self.__uplink_queue)
                            device.start()
                    self.__connected = connected_devices == len(self.__devices)
                    if not self.__uplink_queue.empty():
                        data = self.__uplink_queue.get()
                        self.__gateway.send_to_storage(self.name, self.__id, data)
                    else:
                        sleep(0.05)
                except Exception as e:
                    self._log.error("Failed to process data from device %s, error: %s", self.name, e)
        except Exception as e:
            self._log.error("Failed to process data from device %s, error: %s", self.name, e)

    def close(self):
        """
        Service method to stop the connector and all devices connected to it.
        """
        self.stopped.set()
        for device in self.__devices:
            self.__gateway.del_device(device.name)
            device.stop()
        self._log.stop()

    def on_attributes_update(self, content):
        """
        Callback method to process attribute updates from the platform.
        """
        self._log.debug("Received attribute update: %s", content)
        device_name = content.get("device")
        if device_name is not None:
            for device in self.__devices:
                if device_name == device.name:
                    request_config = device.config.get("attributeUpdates")
                    if request_config is not None:
                        attribute_config_found = False
                        for attribute_config in request_config:
                            attribute = attribute_config.get("attributeOnPlatform")
                            if attribute is not None and attribute in content["data"]:
                                attribute_config_found = True
                                try:
                                    value = content["data"][attribute]
                                    str_to_send = str(attribute_config["stringToDevice"]
                                                      .replace("${" + attribute + "}", str(value))
                                                      .replace("${deviceName}", device_name)
                                                      .replace("${deviceType}", device.type)
                                                      ).encode("UTF-8")
                                    device.write(str_to_send)
                                except Exception as e:
                                    self._log.error("Failed to send attribute update to device %s: %s",
                                                    device_name, e)
                        if not attribute_config_found:
                            self._log.error("Attribute update configuration for key %s for device %s not found",
                                            list(content['data'].keys())[0], device_name)
                    else:
                        self._log.error("Attribute update configuration for device %s not found", device_name)
        else:
            self._log.error("Device name is not provided in the attribute update request: %s", content)

    def server_side_rpc_handler(self, content):
        """
        Callback method to process RPC requests from the platform.
        """
        self._log.debug("Received RPC request: %s", content)
        device_name = content.get("device")
        rpc_data = content.get("data", {})
        rpc_method = rpc_data.get("method")
        req_id = rpc_data.get("id")
        params = rpc_data.get("params")
        if device_name is not None:
            for device in self.__devices:
                if device_name == device.name:
                    result = device.handle_rpc_request(rpc_method, params)
                    if "error" in result:
                        self._log.error("Failed to process RPC request for device %s, error: %s",
                                        device_name, result["error"])
                    if result is not None:
                        self.__gateway.send_rpc_reply(device=device_name,
                                                      req_id=req_id,
                                                      content=result,
                                                      wait_for_publish=True,
                                                      quality_of_service=1)
        else:
            self._log.error("Device name is not provided in the RPC request: %s", content)



#      Copyright 2025. ThingsBoard
#
#      Licensed under the Apache License, Version 2.0 (the "License");
#      you may not use this file except in compliance with the License.
#      You may obtain a copy of the License at
#
#          http://www.apache.org/licenses/LICENSE-2.0
#
#      Unless required by applicable law or agreed to in writing, software
#      distributed under the License is distributed on an "AS IS" BASIS,
#      WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#      See the License for the specific language governing permissions and
#      limitations under the License.

from threading import Thread
from time import monotonic, sleep

from thingsboard_gateway.connectors.knx.knx_uplink_converter import KNXUplinkConverter
from thingsboard_gateway.gateway.constants import CONVERTER_PARAMETER, UPLINK_PREFIX
from thingsboard_gateway.tb_utility.tb_loader import TBModuleLoader
from thingsboard_gateway.tb_utility.tb_utility import TBUtility


class Device(Thread):
    def __init__(self, connector_type, is_client_connected, config, process_request_queue, logger):
        super().__init__()

        self.__log = logger
        self.__connector_type = connector_type
        self.__is_client_connected = is_client_connected
        self.__config = config

        self.group_addresses_to_read = {}
        self.__fill_group_addresses_to_read()

        device_info = self.__config['deviceInfo']
        self.name = device_info.get('deviceNameExpression', 'KNX Device')
        self.daemon = True
        self.__stopped = False

        self.uplink_converter = self.__load_uplink_converter()

        self.__process_request_queue = process_request_queue

        self.__poll_period = self.__config.get('pollPeriod', 10000) / 1000

        self.__last_poll_time = 0

        self.start()

    def __str__(self):
        return f'Device({self.name})'

    def __fill_group_addresses_to_read(self):
        """
        Fill self.group_addresses_to_read with all group addresses
        to read from the device from:
        - attributes
        - temeseries
        - deviceNameExpression
        - deviceTypeExpression

        Data have the following structure:
        {
            "1.0.5": {
                "type": "string",
                "keys": ["vendor", "deviceName"]
            }
        }
        """

        self.__fill_group_addresses_from_device_info()

        for section in ('attributes', 'timeseries'):
            for datapoint_config in self.__config.get(section, []):
                self.add_group_address(datapoint_config['groupAddress'],
                                       datapoint_config.get('type'),
                                       datapoint_config['key'])

    def add_group_address(self, group_address, datatype, key):
        if not self.group_addresses_to_read.get(group_address):
            self.group_addresses_to_read[group_address] = {
                'type': datatype,
                'keys': []
            }

        self.group_addresses_to_read[group_address]['keys'].append(key)

    def __fill_group_addresses_from_device_info(self):
        device_info = self.__config['deviceInfo']

        device_name_group_address = self.__get_group_address_from_device_name(device_info)
        if device_name_group_address:
            self.add_group_address(device_name_group_address,
                                   device_info.get('deviceNameDataType'),
                                   'deviceName')

        device_type_group_address = self.__get_group_address_from_device_profile(device_info)
        if device_type_group_address:
            self.add_group_address(device_type_group_address,
                                   device_info.get('deviceProfileDataType'),
                                   'deviceType')

    @staticmethod
    def __get_group_address_from_device_name(device_info):
        if device_info.get('deviceNameExpressionSource') == 'expression':
            if '${' in device_info['deviceNameExpression']:
                expression = Device.__get_group_address_from_expression(device_info['deviceNameExpression'])

                return expression

    @staticmethod
    def __get_group_address_from_device_profile(device_info):
        if device_info.get('deviceProfileExpressionSource') == 'expression':
            if '${' in device_info['deviceProfileExpression']:
                expression = Device.__get_group_address_from_expression(device_info['deviceProfileExpression'])

                return expression

    @staticmethod
    def __get_group_address_from_expression(expression):
        return TBUtility.get_value(expression, get_tag=True)

    def __load_uplink_converter(self):
        try:
            if self.__config.get(UPLINK_PREFIX + CONVERTER_PARAMETER) is not None:
                converter_class = TBModuleLoader.import_module(self.__connector_type,
                                                               self.__config[UPLINK_PREFIX + CONVERTER_PARAMETER])
                converter = converter_class(self.uplink_converter_config, self.__log)
            else:
                converter = KNXUplinkConverter(self.__config, self.__log)

            return converter
        except Exception as e:
            self.__log.error('Failed to load uplink converter for % device: %s', self.name, e)

    def run(self):
        while not self.__stopped:
            if monotonic() - self.__last_poll_time > self.__poll_period and self.__is_client_connected.is_set():
                self.__put_to_queue()
                self.__last_poll_time = monotonic()
            else:
                sleep(.01)

    def stop(self):
        self.__stopped = True

    def __put_to_queue(self):
        try:
            self.__process_request_queue.put(self)
        except Exception as e:
            self.__log.exception(e)



#     Copyright 2025. ThingsBoard
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

from pymodbus.constants import Endian


class BytesUplinkConverterConfig:
    def __init__(self, **kwargs):
        self.report_strategy = kwargs.get('reportStrategy')
        self.device_name = kwargs['deviceName']
        self.device_type = kwargs.get('deviceType', 'default')
        self.byte_order = Endian.BIG if kwargs.get('byteOrder', 'LITTLE').upper() == "BIG" else Endian.LITTLE
        self.word_order = Endian.BIG if kwargs.get('wordOrder', 'LITTLE').upper() == "BIG" else Endian.LITTLE
        self.telemetry = kwargs.get('timeseries', [])
        self.attributes = kwargs.get('attributes', [])

    def is_readable(self):
        return len(self.telemetry) > 0 or len(self.attributes) > 0



#     Copyright 2025. ThingsBoard
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

import datetime
from getpass import getuser
from logging import getLogger, setLoggerClass
from os import environ
from platform import system as platform_system
from re import search, findall
from time import monotonic, sleep
from typing import Union, TYPE_CHECKING
from uuid import uuid4

from cachetools import TTLCache
from cryptography import x509
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.x509.oid import NameOID
from dateutil import parser
from jsonpath_rw import parse
from orjson import JSONDecodeError, dumps, loads, OPT_NON_STR_KEYS

from thingsboard_gateway.gateway.constants import SECURITY_VAR, REPORT_STRATEGY_PARAMETER
from thingsboard_gateway.gateway.entities.datapoint_key import DatapointKey
from thingsboard_gateway.gateway.entities.report_strategy_config import ReportStrategyConfig
from thingsboard_gateway.tb_utility.tb_logger import TbLogger

if TYPE_CHECKING:
    from thingsboard_gateway.gateway.entities.converted_data import ConvertedData

setLoggerClass(TbLogger)
log = getLogger("service")


class TBUtility:
    JSONPATH_EXPRESSION_CACHE = TTLCache(maxsize=10000, ttl=30)

    # Data conversion methods

    @staticmethod
    def decode(message, return_raw=False):
        if return_raw:
            return getattr(message, 'payload', message)

        payload = message if isinstance(message, bytes) else getattr(message, 'payload', None)
        if payload is None:
            return message

        if isinstance(payload, bytes):
            try:
                return loads(payload.decode("utf-8", "ignore"))
            except JSONDecodeError:
                try:
                    return payload.decode("utf-8")
                except UnicodeDecodeError:
                    return payload
        else:
            try:
                return loads(payload)
            except JSONDecodeError:
                return payload

    @staticmethod
    def validate_converted_data(data: Union[dict, 'ConvertedData']):
        from thingsboard_gateway.gateway.entities.converted_data import ConvertedData

        errors = []
        if isinstance(data, ConvertedData):
            if data.device_name is None:
                errors.append('deviceName is empty')
            if not data.telemetry and not data.attributes:
                errors.append('No telemetry and attributes')
        else:
            if not data.get('deviceName'):
                errors.append('deviceName is empty')

            got_attributes = False
            got_telemetry = False

            if data.get("attributes") is not None and len(data.get("attributes")) > 0:
                got_attributes = True

            if data.get("telemetry") is not None:
                for entry in data.get("telemetry"):
                    if (entry.get("ts") is not None and len(entry.get("values")) > 0) or entry.get("ts") is None:
                        got_telemetry = True
                        break

            if got_attributes is False and got_telemetry is False:
                errors.append('No telemetry and attributes')

        if errors:
            json_data = dumps(data.to_dict(), option=OPT_NON_STR_KEYS) if isinstance(data, ConvertedData) else dumps(data, option=OPT_NON_STR_KEYS)
            if isinstance(json_data, bytes):
                log.error("Found errors: " + str(errors) + " in data: " + json_data.decode("UTF-8"))
            else:
                log.error("Found errors: " + str(errors) + " in data: " + json_data)
            return False
        return True

    @staticmethod
    def topic_to_regex(topic):
        return topic.replace("+", "[^/]+").replace("#", ".+").replace('$', '\\$')

    @staticmethod
    def regex_to_topic(regex):
        return regex.replace("[^/]+", "+").replace(".+", "#").replace('\\$', '$')

    @staticmethod
    def get_value(expression, body=None, value_type="string", get_tag=False, expression_instead_none=False):
        if isinstance(body, str):
            body = loads(body)
        if not expression:
            return ''
        positions = search(r'\${(?:(.*))}', expression)
        if positions is not None:
            p1 = positions.regs[-1][0]
            p2 = positions.regs[-1][1]
        else:
            p1 = 0
            p2 = len(expression)
        target_str = str(expression[p1:p2])
        if get_tag:
            return target_str
        full_value = None
        try:
            if isinstance(body, dict) and target_str.split()[0] in body:
                if value_type.lower() == "string":
                    full_value = (str(expression[0: max(p1 - 2, 0)]) +
                                  str(body[target_str.split()[0]]) +
                                  str(expression[p2 + 1:len(expression)]))
                else:
                    full_value = body.get(target_str.split()[0])
            elif isinstance(body, (dict, list)):
                try:
                    if " " in target_str:
                        target_str = '.'.join('"' + section_key + '"' if " " in section_key else section_key for section_key in target_str.split('.'))  # noqa
                    jsonpath_expression = TBUtility.JSONPATH_EXPRESSION_CACHE.get(target_str)
                    if jsonpath_expression is None:
                        jsonpath_expression = parse(target_str)
                        TBUtility.JSONPATH_EXPRESSION_CACHE[target_str] = jsonpath_expression
                    jsonpath_match = jsonpath_expression.find(body)
                    if jsonpath_match:
                        full_value = jsonpath_match[0].value
                except Exception as e:
                    log.debug(e)
            elif isinstance(body, (str, bytes)):
                search_result = search(expression, body)
                if search_result.groups():
                    full_value = search_result.group(0)
            if expression_instead_none and full_value is None:
                full_value = expression
        except Exception as e:
            log.exception(e)
        return full_value

    @staticmethod
    def get_values(expression, body=None, value_type="string", get_tag=False, expression_instead_none=False):
        expression_arr = findall(r'\$\{[${A-Za-z0-9. ^\]\[*_:"-]*\}', expression)

        values = [TBUtility.get_value(exp, body, value_type=value_type, get_tag=get_tag,
                                      expression_instead_none=expression_instead_none) for exp in expression_arr]

        if '${' not in expression:
            values.append(expression)

        return values

    @staticmethod
    def replace_params_tags(text, data):
        if '${' in text:
            for item in text.split('/'):
                if '${' in item:
                    tag = '${' + TBUtility.get_value(item, data['data'], 'params', get_tag=True) + '}'
                    value = TBUtility.get_value(item, data['data'], 'params', expression_instead_none=True)
                    text = text.replace(tag, str(value))
        return text

    @staticmethod
    def get_dict_key_by_value(dictionary: dict, value):
        try:
            return next(key for key, val in dictionary.items() if val == value)
        except StopIteration:
            return None

    @staticmethod
    def str_to_bool(val) -> bool:
        if isinstance(val, bool):
            return val

        if isinstance(val, int):
            if val in (0, 1):
                return bool(val)
            raise ValueError(f"Invalid truth value (int): {val!r}")

        try:
            s = str(val).strip().lower()
        except Exception:
            raise ValueError(f"Cannot convert {val!r} to boolean")

        true_vals = {"y", "yes", "t", "true", "on", "1"}
        false_vals = {"n", "no", "f", "false", "off", "0"}

        if s in true_vals:
            return True
        if s in false_vals:
            return False

        raise ValueError(f"Invalid truth value: {val!r}")

    @staticmethod
    def convert_data_type(data, new_type, use_eval=False):
        current_type = type(data)
        # use 'in' check instead of equality for such case like 'str' and 'string'
        new_type = new_type.lower()
        if current_type.__name__ in new_type:
            return data

        evaluated_data = eval(data, globals(), {}) if use_eval else data
        try:
            if 'int' in new_type or 'long' in new_type:
                return int(float(evaluated_data))
            elif 'float' == new_type or 'double' == new_type:
                return float(evaluated_data)
            elif 'bool' in new_type:
                return TBUtility.str_to_bool(evaluated_data)
            else:
                return str(evaluated_data)
        except ValueError:
            return str(evaluated_data)

    @staticmethod
    def convert_key_to_datapoint_key(key, device_report_strategy, key_config, logger=None):
        key_report_strategy = None
        if device_report_strategy is not None:
            key_report_strategy = device_report_strategy
        if key_config.get(REPORT_STRATEGY_PARAMETER) is not None:
            try:
                key_report_strategy = ReportStrategyConfig(key_config.get(REPORT_STRATEGY_PARAMETER))
            except ValueError:
                if logger is not None:
                    logger.trace("Report strategy config is not specified for key %s", key)
        return DatapointKey(key, key_report_strategy)

    # Service methods

    @staticmethod
    def install_package(package, version="upgrade", force_install=False):
        from sys import executable, prefix, base_prefix, modules
        from subprocess import check_call
        import site
        from importlib import reload

        result = False
        installation_sign = "==" if ">=" not in version else ""

        if prefix != base_prefix:
            if force_install:
                result = check_call([executable, '-m', 'pip', 'install', package + '==' + version, '--force-reinstall'])
            elif version.lower() == "upgrade":
                result = check_call([executable, "-m", "pip", "install", package, "--upgrade"])
            else:
                if TBUtility.get_package_version(package) is None:
                    result = check_call([executable, "-m", "pip", "install", package + installation_sign + version])
        else:
            if force_install:
                result = check_call(
                    [executable, '-m', 'pip', 'install', package + '==' + version, '--force-reinstall', "--user"])
            elif version.lower() == "upgrade":
                result = check_call([executable, "-m", "pip", "install", package, "--upgrade", "--user"])
            else:
                if TBUtility.get_package_version(package) is None:
                    result = check_call(
                        [executable, "-m", "pip", "install", package + installation_sign + version, "--user"])

        # Because `pip` is running in a subprocess the newly installed modules and libraries are
        # not immediately available to the current runtime.
        # Refreshing sys.path fixes this. See:
        # https://stackoverflow.com/questions/4271494/what-sets-up-sys-path-with-python-and-when
        reload(site)
        reload(site)
        if package in modules:
            del modules[package]

        return result

    @staticmethod
    def get_package_version(package):
        from pkg_resources import get_distribution
        current_package_version = None
        try:
            current_package_version = get_distribution(package)
        except Exception:
            pass
        return current_package_version

    @staticmethod
    def get_or_create_connector_id(connector_conf):
        connector_id = str(uuid4())
        if isinstance(connector_conf, dict):
            if connector_conf.get('id') is not None:
                connector_id = connector_conf['id']
        elif isinstance(connector_conf, str):
            start_find = connector_conf.find("{id_var_start}")
            end_find = connector_conf.find("{id_var_end}")
            if start_find > -1 and end_find > -1:
                connector_id = connector_conf[start_find + 13:end_find]
        return connector_id

    @staticmethod
    def generate_certificate(old_certificate_path, old_key_path, old_certificate=None):
        key = ec.generate_private_key(ec.SECP256R1())
        public_key = key.public_key()
        builder = x509.CertificateBuilder()
        builder = builder.subject_name(old_certificate.subject if old_certificate else x509.Name(
            [x509.NameAttribute(NameOID.COMMON_NAME, u'localhost'), ]))
        builder = builder.issuer_name(old_certificate.issuer if old_certificate else x509.Name(
            [x509.NameAttribute(NameOID.COMMON_NAME, u'localhost'), ]))
        builder = builder.not_valid_before(datetime.datetime.today() - datetime.timedelta(days=1))
        builder = builder.not_valid_after(datetime.datetime.today() + (datetime.timedelta(1, 0, 0) * 365))
        builder = builder.serial_number(x509.random_serial_number())
        builder = builder.public_key(public_key)
        certificate = builder.sign(private_key=key, algorithm=hashes.SHA256())

        cert = certificate.public_bytes(serialization.Encoding.PEM)
        with open(old_certificate_path, 'wb+') as f:
            f.write(cert)

        key = key.private_bytes(encoding=serialization.Encoding.PEM,
                                format=serialization.PrivateFormat.TraditionalOpenSSL,
                                encryption_algorithm=serialization.NoEncryption())
        with open(old_key_path, 'wb+') as f:
            f.write(key)

        return cert

    @staticmethod
    def check_certificate(certificate, key=None, generate_new=True, days_left=3):
        cert_detail = x509.load_pem_x509_certificate(open(certificate, 'rb').read())

        if cert_detail.not_valid_after - datetime.datetime.now() <= datetime.timedelta(days=days_left):
            if generate_new:
                return TBUtility.generate_certificate(certificate, key, cert_detail)
            else:
                return True

    @staticmethod
    def get_data_size(data):
        return len(dumps(data, option=OPT_NON_STR_KEYS))

    @staticmethod
    def update_main_config_with_env_variables(config):
        env_variables = TBUtility.get_service_environmental_variables()
        config['thingsboard'] = {**config['thingsboard'], **env_variables}
        return config

    @staticmethod
    def resolve_different_ts_formats(data: dict, config: dict, logger, default_ts: bool = True):
        ts_field_expression = config.get('tsField')
        if ts_field_expression is not None:
            ts_field_key = None
            try:
                ts_field_key = TBUtility.get_value(ts_field_expression, data, get_tag=True)

                if data.get(ts_field_key) is not None:
                    parsed_configuration_data = parser.parse(data[ts_field_key], dayfirst=config.get('dayfirst', False), yearfirst=config.get('yearfirst', False))
                    return int(parsed_configuration_data.timestamp() * 1000)
                return data.get(ts_field_expression)


            except Exception as e:
                logger.debug("Error while parsing timestamp %s: %s with configured tsField: %s",
                             ts_field_key, e, config['tsField'])
        return data.get('ts', data.get('timestamp')) if default_ts else None

    @staticmethod
    def get_service_environmental_variables():
        env_variables = {
            'host': environ.get('host'),
            'port': int(environ.get('port')) if environ.get('port') else None,
            'type': environ.get('type'),
            'accessToken': environ.get('accessToken'),
            'caCert': environ.get('caCert'),
            'privateKey': environ.get('privateKey'),
            'cert': environ.get('cert'),
            'clientId': environ.get('clientId'),
            'password': environ.get('password'),
            'proxy_host': environ.get('proxy_host'),
            'proxy_port': int(environ.get('proxy_port')) if environ.get('proxy_port') else None
        }

        if platform_system() != 'Windows':
            env_variables['username'] = environ.get('username')
        elif environ.get('username') is not None and getuser().lower() != environ.get('username').lower():
            env_variables['username'] = environ.get('username')
        if environ.get('TB_GW_HOST'):
            env_variables['host'] = environ.get('TB_GW_HOST')
        if environ.get('TB_GW_PORT'):
            env_variables['port'] = int(environ.get('TB_GW_PORT'))
        if environ.get('TB_GW_PROXY_HOST'):
            env_variables['proxy_host'] = environ.get('TB_GW_PROXY_HOST')
        if environ.get('TB_GW_PROXY_PORT'):
            env_variables['proxy_port'] = int(environ.get('TB_GW_PROXY_PORT'))
        if environ.get('TB_GW_SECURITY_TYPE'):
            env_variables['type'] = environ.get('TB_GW_SECURITY_TYPE')
        if environ.get('TB_GW_ACCESS_TOKEN'):
            env_variables['accessToken'] = environ.get('TB_GW_ACCESS_TOKEN')
        if environ.get('TB_GW_CA_CERT'):
            env_variables['caCert'] = environ.get('TB_GW_CA_CERT')
        if environ.get('TB_GW_PRIVATE_KEY'):
            env_variables['privateKey'] = environ.get('TB_GW_PRIVATE_KEY')
        if environ.get('TB_GW_CERT'):
            env_variables['cert'] = environ.get('TB_GW_CERT')
        if environ.get('TB_GW_CLIENT_ID'):
            env_variables['clientId'] = environ.get('TB_GW_CLIENT_ID')
        if environ.get('TB_GW_USERNAME'):
            env_variables['username'] = environ.get('TB_GW_USERNAME')
        if environ.get('TB_GW_PASSWORD'):
            env_variables['password'] = environ.get('TB_GW_PASSWORD')

        if environ.get('TB_GW_RATE_LIMITS'):
            env_variables['rateLimits'] = environ.get('TB_GW_RATE_LIMITS')
            env_variables['messagesRateLimits'] = environ.get('TB_GW_RATE_LIMITS')
            env_variables['deviceMessagesRateLimits'] = environ.get('TB_GW_RATE_LIMITS')
            env_variables['deviceRateLimits'] = environ.get('TB_GW_RATE_LIMITS')
        if environ.get('TB_GW_DP_RATE_LIMITS'):
            env_variables['dpRateLimits'] = environ.get('TB_GW_DP_RATE_LIMITS')
            env_variables['deviceDpRateLimits'] = environ.get('TB_GW_DP_RATE_LIMITS')

        if environ.get('TB_GW_MESSAGES_RATE_LIMITS'):
            env_variables['messagesRateLimits'] = environ.get('TB_GW_MESSAGES_RATE_LIMITS')
        if environ.get('TB_GW_DEVICE_MESSAGES_RATE_LIMIT'):
            env_variables['deviceMessagesRateLimits'] = environ.get('TB_GW_DEVICE_MESSAGES_RATE_LIMIT')
        if environ.get('TB_GW_DEVICE_RATE_LIMITS'):
            env_variables['deviceRateLimits'] = environ.get('TB_GW_DEVICE_RATE_LIMITS')
        if environ.get('TB_GW_DEVICE_DP_RATE_LIMITS'):
            env_variables['deviceDpRateLimits'] = environ.get('TB_GW_DEVICE_DP_RATE_LIMITS')

        converted_env_variables = {}

        for (key, value) in env_variables.items():
            if value is not None:
                if key in SECURITY_VAR:
                    if not converted_env_variables.get('security'):
                        converted_env_variables['security'] = {}

                    converted_env_variables['security'][key] = value
                else:
                    converted_env_variables[key] = value

        return converted_env_variables

    @staticmethod
    def get_provisioning_configuration_from_envs() -> dict:
        provision_request = {}
        provisioning_device_key = environ.get('TB_GW_PROVISIONING_DEVICE_KEY')
        provisioning_device_secret = environ.get('TB_GW_PROVISIONING_DEVICE_SECRET')
        if provisioning_device_key and provisioning_device_secret:
            provision_request['provisionDeviceKey'] = provisioning_device_key
            provision_request['provisionDeviceSecret'] = provisioning_device_secret

            provisioning_device_name = environ.get('TB_GW_PROVISIONING_DEVICE_NAME')
            provisioning_device_access_token = environ.get('TB_GW_PROVISIONING_DEVICE_ACCESS_TOKEN')
            provisioning_device_username = environ.get('TB_GW_PROVISIONING_DEVICE_USERNAME')
            provisioning_device_password = environ.get('TB_GW_PROVISIONING_DEVICE_PASSWORD')
            provisioning_device_client_id = environ.get('TB_GW_PROVISIONING_DEVICE_CLIENT_ID')
            provisioning_device_ca_cert_filename = environ.get('TB_GW_PROVISIONING_DEVICE_CA_CERT_FILENAME')

            if provisioning_device_name is not None:
                provision_request["deviceName"] = provisioning_device_name

            if provisioning_device_access_token is not None:
                provision_request["access_token"] = provisioning_device_access_token
                provision_request["type"] = "ACCESS_TOKEN"
            elif (provisioning_device_username is not None
                  or provisioning_device_password is not None
                  or provisioning_device_client_id is not None):
                provision_request["username"] = provisioning_device_username
                provision_request["password"] = provisioning_device_password
                provision_request["client_id"] = provisioning_device_client_id
                provision_request["type"] = "MQTT_BASIC"
            elif provisioning_device_ca_cert_filename is not None:
                provision_request["caCert"] = provisioning_device_ca_cert_filename
                provision_request["type"] = "X509_CERTIFICATE"
        else:
            log.debug("Provisioning device key or secret is not set. Skipping provisioning credentials.")
        return provision_request

    @staticmethod
    def while_thread_alive(thread, timeout=10) -> bool:
        start_time = monotonic()

        while thread.is_alive():
            if monotonic() - start_time > timeout:
                return True

            sleep(.1)

        return False

    @staticmethod
    def kebab_case_to_camel_case(snake_str: str) -> str:
        components = snake_str.split('-')
        return components[0] + ''.join(x.title() for x in components[1:]) if len(components) > 1 else components[0]



#     Copyright 2025. ThingsBoard
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

from thingsboard_gateway.connectors.socket.socket_uplink_converter import SocketUplinkConverter
from thingsboard_gateway.gateway.constants import REPORT_STRATEGY_PARAMETER
from thingsboard_gateway.gateway.entities.converted_data import ConvertedData
from thingsboard_gateway.gateway.entities.report_strategy_config import ReportStrategyConfig
from thingsboard_gateway.gateway.entities.telemetry_entry import TelemetryEntry
from thingsboard_gateway.gateway.statistics.statistics_service import StatisticsService
from thingsboard_gateway.gateway.statistics.decorators import CollectStatistics
from thingsboard_gateway.tb_utility.tb_utility import TBUtility


class BytesSocketUplinkConverter(SocketUplinkConverter):
    def __init__(self, config, logger):
        self._log = logger
        self.__config = config

    @CollectStatistics(start_stat_type='receivedBytesFromDevices',
                       end_stat_type='convertedBytesFromDevice')
    def convert(self, config, data):
        converted_data = ConvertedData(device_name=self.__config['deviceName'],
                                       device_type=self.__config['deviceType'])

        if data is None:
            return converted_data

        device_report_strategy = None
        try:
            device_report_strategy = ReportStrategyConfig(config.get(REPORT_STRATEGY_PARAMETER))
        except ValueError as e:
            self._log.trace("Report strategy config is not specified for device %s: %s", self.__config['deviceName'], e)

        try:
            for section in ('telemetry', 'attributes'):
                for item in config[section]:
                    try:
                        byte_from = item.get('byteFrom')
                        byte_to = item.get('byteTo')

                        byte_to = byte_to if byte_to != -1 else len(data)
                        decoded_data = data[byte_from:byte_to]
                        if config['encoding'] == 'hex':
                            decoded_data = decoded_data.hex()
                        else:
                            try:
                                decoded_data = decoded_data.replace(b"\x00", b'').decode(config['encoding'])
                            except UnicodeDecodeError:
                                decoded_data = str(decoded_data)

                        if item.get('key') is not None:
                            datapoint_key = TBUtility.convert_key_to_datapoint_key(item['key'], device_report_strategy,
                                                                                   item, self._log)
                            if section == 'attributes':
                                converted_data.add_to_attributes(datapoint_key, decoded_data)
                            else:
                                telemetry_entry = TelemetryEntry({datapoint_key: decoded_data})
                                converted_data.add_to_telemetry(telemetry_entry)
                        else:
                            self._log.error('Key for %s not found in config: %s', config['type'], config['section_config'])
                    except Exception as e:
                        self._log.exception(e)
        except Exception as e:
            StatisticsService.count_connector_message(self._log.name, 'convertersMsgDropped')
            self._log.exception(e)

        self._log.debug("Converted data: %s", converted_data)
        StatisticsService.count_connector_message(self._log.name, 'convertersAttrProduced',
                                                  count=converted_data.attributes_datapoints_count)
        StatisticsService.count_connector_message(self._log.name, 'convertersTsProduced',
                                                  count=converted_data.telemetry_datapoints_count)

        return converted_data



#     Copyright 2025. ThingsBoard
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.


from simplejson import dumps

from thingsboard_gateway.connectors.rest.rest_converter import RESTConverter
from thingsboard_gateway.gateway.constants import REPORT_STRATEGY_PARAMETER
from thingsboard_gateway.gateway.entities.converted_data import ConvertedData
from thingsboard_gateway.gateway.entities.report_strategy_config import ReportStrategyConfig
from thingsboard_gateway.gateway.entities.telemetry_entry import TelemetryEntry
from thingsboard_gateway.gateway.statistics.decorators import CollectStatistics
from thingsboard_gateway.tb_utility.tb_utility import TBUtility
from thingsboard_gateway.gateway.statistics.statistics_service import StatisticsService


class JsonRESTUplinkConverter(RESTConverter):

    def __init__(self, config, logger):
        self._log = logger
        self.__config = config

    @CollectStatistics(start_stat_type='receivedBytesFromDevices',
                       end_stat_type='convertedBytesFromDevice')
    def convert(self, config, data):
        datatypes = {"attributes": "attributes",
                     "timeseries": "telemetry"}
        device_name = None
        device_type = None

        try:
            device_info = self.__config.get("deviceInfo")
            if device_info.get("deviceNameExpression") is not None:
                if device_info.get("deviceNameExpressionSource") == "constant":
                    device_name = device_info.deviceNameExpression
                else:
                    device_name_tags = TBUtility.get_values(device_info.get("deviceNameExpression"), data, get_tag=True)
                    device_name_values = TBUtility.get_values(device_info.get("deviceNameExpression"), data,
                                                              expression_instead_none=True)
                    device_name = device_info.get("deviceNameExpression")
                    for (device_name_tag, device_name_value) in zip(device_name_tags, device_name_values):
                        is_valid_key = "${" in device_info.get("deviceNameExpression") and "}" in \
                                       device_info.get("deviceNameExpression")
                        device_name = device_name.replace('${' + str(device_name_tag) + '}',
                                                          str(device_name_value)) \
                            if is_valid_key else device_name_tag
            else:
                self._log.error("The expression for looking \"device name\" not found in config %s",
                                dumps(device_info))

            if device_info.get("deviceProfileExpression") is not None:
                if device_info.get("deviceProfileExpressionSource") == "constant":
                    device_type = device_info.get("deviceProfileExpression")
                else:
                    device_profile_tags = TBUtility.get_values(device_info.get("deviceProfileExpression"), data,
                                                               get_tag=True)
                    device_profile_values = TBUtility.get_values(device_info.get("deviceProfileExpression"), data,
                                                                 expression_instead_none=True)
                    device_type = device_info.get("deviceProfileExpression")
                    for (device_profile_tag, device_profile_value) in zip(device_profile_tags, device_profile_values):
                        is_valid_key = "${" in device_info.get("deviceProfileExpression") and "}" in \
                                       device_info.get("deviceProfileExpression")
                        device_type = device_type.replace('${' + str(device_profile_tag) + '}',
                                                          str(device_profile_value)) \
                            if is_valid_key else device_profile_tag
            else:
                self._log.error("The expression for looking \"device profile\" not found in config %s",
                                dumps(device_info))
        except Exception as e:
            StatisticsService.count_connector_message(self._log.name, 'convertersMsgDropped')
            self._log.error('Error in converter, for config: \n%s\n and message: \n%s\n %s', dumps(self.__config), data,
                            e)

        converted_data = ConvertedData(device_name=device_name, device_type=device_type)
        device_report_strategy = None
        try:
            device_report_strategy = ReportStrategyConfig(self.__config.get(REPORT_STRATEGY_PARAMETER))
        except ValueError as e:
            self._log.trace("Report strategy config is not specified for device %s: %s", device_name, e)

        try:
            for datatype in datatypes:
                for datatype_config in self.__config.get(datatype, []):
                    values = TBUtility.get_values(datatype_config["value"], data, datatype_config["type"],
                                                  expression_instead_none=False)
                    values_tags = TBUtility.get_values(datatype_config["value"], data, datatype_config["type"],
                                                       get_tag=True)

                    keys = TBUtility.get_values(datatype_config["key"], data, datatype_config["type"],
                                                expression_instead_none=False)
                    keys_tags = TBUtility.get_values(datatype_config["key"], data, get_tag=True)

                    full_key = datatype_config['key']
                    for (key_tag, key) in zip(keys_tags, keys):
                        is_valid_key = "${" in datatype_config["key"] and "}" in \
                                       datatype_config["key"]
                        full_key = full_key.replace('${' + str(key_tag) + '}',
                                                    str(key)) if is_valid_key else key

                    full_value = datatype_config['value']
                    for (value_tag, value) in zip(values_tags, values):
                        is_valid_value = "${" in datatype_config["value"] and "}" in \
                                         datatype_config["value"]

                        full_value = full_value.replace('${' + str(value_tag) + '}',
                                                        str(value)) if is_valid_value else value

                    if full_key != 'None' and full_value != 'None':
                        datapoint_key = TBUtility.convert_key_to_datapoint_key(full_key, device_report_strategy,
                                                                               datatype_config, self._log)
                        if datatype == 'attributes':
                            converted_data.add_to_attributes(datapoint_key, full_value)
                        else:
                            ts = TBUtility.resolve_different_ts_formats(data=data, config=datatype_config, logger=self._log)
                            telemetry_entry = TelemetryEntry({datapoint_key: full_value}, ts)
                            converted_data.add_to_telemetry(telemetry_entry)
        except Exception as e:
            StatisticsService.count_connector_message(self._log.name, 'convertersMsgDropped')
            self._log.error('Error in converter, for config: \n%s\n and message: \n%s\n %s', dumps(self.__config),
                            str(data), e)
        self._log.info("Converted data: %s", converted_data)

        StatisticsService.count_connector_message(self._log.name, 'convertersAttrProduced',
                                                  count=converted_data.attributes_datapoints_count)
        StatisticsService.count_connector_message(self._log.name, 'convertersTsProduced',
                                                  count=converted_data.telemetry_datapoints_count)

        return converted_data



from dataclasses import dataclass
from typing import List


@dataclass
class Device:
    jid: str
    device_name_expression: str
    device_type_expression: str
    attributes: List
    timeseries: List
    attribute_updates: List
    server_side_rpc: List
    converter = None

    def set_converter(self, converter):
        self.converter = converter



#     Copyright 2025. ThingsBoard
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

from csv import reader as csv_reader
from base64 import b64decode
from io import StringIO

from thingsboard_gateway.connectors.bacnet.constants import (
    DEVICE_OBJECT_TYPE_CODE,
    EDE_CONFIG_PARAMETER,
    EDE_FILE_PATH_PARAMETER,
    OBJECT_TYPE_PARAMETER,
    OBJECT_INSTANCE_PARAMETER,
    DEVICE_OBJ_INSTANCE_PARAMETER,
    OBJECT_NAME_PARAMETER
)
from thingsboard_gateway.connectors.bacnet.entities.ede_connector_config import EDEConnectorConfig


class EDEParser:
    @staticmethod
    def is_ede_config(config) -> bool:
        return EDE_FILE_PATH_PARAMETER in config or EDE_CONFIG_PARAMETER in config

    @staticmethod
    def parse(config):
        if EDE_FILE_PATH_PARAMETER in config:
            parsed_config = EDEParser.parse_from_file(config[EDE_FILE_PATH_PARAMETER])
            config.pop(EDE_FILE_PATH_PARAMETER, None)
        elif EDE_CONFIG_PARAMETER in config:
            parsed_config = EDEParser.parse_from_base64(config[EDE_CONFIG_PARAMETER])
            config.pop(EDE_CONFIG_PARAMETER, None)

        return {**config, **parsed_config.config}

    @staticmethod
    def parse_from_file(file_path: str) -> str:
        with open(file_path, newline='', encoding='utf-8') as file:
            reader = csv_reader(file, delimiter=';')
            data = [row for row in reader]

        return EDEParser.__parse(data)

    @staticmethod
    def parse_from_base64(base64_string) -> str:
        decoded_data = b64decode(base64_string)
        file = StringIO(decoded_data.decode('utf-8'))
        data = csv_reader(file)

        return EDEParser.__parse(data)

    @staticmethod
    def __parse(data):
        csv_data = EDEParser.__parse_csv(data)
        return EDEParser.__parse_ede(csv_data)

    @staticmethod
    def __parse_csv(data):
        headers = []
        output = []

        for row in data:
            if not row or all(cell.strip() == '' for cell in row):
                continue

            if not row[0].startswith('#') and len(row) == 2:
                continue

            if row[0].startswith('# keyname'):
                headers = [col.strip().lstrip('#').strip() for col in row]
                continue

            if row[0].startswith('#'):
                continue

            if headers:
                entry = {headers[i]: row[i].strip() if i < len(row) else '' for i in range(len(headers))}
                output.append(entry)

        return output

    @staticmethod
    def __parse_ede(csv_data):
        connector_config = EDEConnectorConfig()

        for row in csv_data:
            if row.get(OBJECT_TYPE_PARAMETER) == DEVICE_OBJECT_TYPE_CODE:
                try:
                    device_objects = EDEParser.__find_device_objects(row[OBJECT_INSTANCE_PARAMETER], csv_data)
                    if len(device_objects) == 0:
                        continue

                    connector_config.add_device(device_name=row.get(OBJECT_NAME_PARAMETER, 'Unknown Device'),
                                                device_id=row[OBJECT_INSTANCE_PARAMETER],
                                                device_objects=device_objects)
                except KeyError:
                    continue

        return connector_config

    @staticmethod
    def __find_device_objects(device_id, data):
        return list(filter(
            lambda row: row.get(DEVICE_OBJ_INSTANCE_PARAMETER) == device_id and
            row.get(OBJECT_TYPE_PARAMETER) != DEVICE_OBJECT_TYPE_CODE,
            data
        ))
